###################################################################
# This file is part of tk120, a utility program for the
# Yaesu VR-120 receiver.
# 
#    Copyright (C) 2001 - 2003, Bob Parnass
# 
# tk120 is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; either version 2 of the License,
# or (at your option) any later version.
# 
# tk120 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with tk120; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# 02111-1307  USA
###################################################################

set Nmessages		231
set BytesPerMessage	32
set NBanks		10
set NChanPerBank	64
set VNChanPerBank	100
set ChanNumberRepeat	no
set HasLabels		yes

###################################################################
# Starting address (in hexadecimal) for each field in
# the memory image.
###################################################################

set ImageAddr(MemoryFreqs)		00  ;# to A00
set ImageAddr(MemoryModes)		03
set ImageAddr(MemoryLabels)		A00  ; # to 18FA
set ImageAddr(SkipFreqs)		1900 ; # to 19FC
set ImageAddr(SkipModes)		1903
set ImageAddr(VFOFreq)			1A00
set ImageAddr(VFOMode)			1A04
set ImageAddr(VFOStep)			1A05
set ImageAddr(FastTuningStep)		1A06

		# Status at the time the radio
		# was last powered on.

		# Note: The current version of tk120 does not use the
		# Status byte information.
		# It is documented here for future reference.

		# 000 000 vfo, not scanning
		# 010 001 limit search
		# 001 010 memory, not scanning
		# 010 000 memory, scanning
		# 010 011 dual watch scan
set ImageAddr(Status)			1A07
set ImageAddr(SetMenuItem)		1A09

			# bits 6,7 = Lamp
			# 00 = auto, 01 = on, 10 = off
set ImageAddr(Lamp)			1A0A
set ImageAddr(BatterySaver)             1A0B
set ImageAddr(1A0C)			1A0C ;# euro = 02
set ImageAddr(Resume)			1A0D
set ImageAddr(1A0E)			1A0E ;# euro = 02
set ImageAddr(FCounter)			1A0F
set ImageAddr(1A10)			1A10

			# bit 1 = keypad lock off/on
			# bit 4 = current lamp status off/on ?
			# bit 5 = attenuator off/on
			# bit 7 = VFO AUTO step off/on
set ImageAddr(FlagByte3)		1A11

# Flags stored in here.
			# bit 2 = Dual Watch (0) or Priority (1)
			# bit 3 = VFO or Limit search
			# bit 4 = Memory Bank scan
			# bit 5 = Key beep
			# bit 6 = Flex step
			# bit 7 = AM antenna
set ImageAddr(FlagByte0)		1A14

			# bit 1 = Decrease S meter sensitivity
			# bit 2 = S meter beep alarm
			# bit 3 = User port off/on
			# bit 4 = Smart Search single (0), continu (1)
			# bit 5 = FM antenna
set ImageAddr(FlagByte1)		1A15

set ImageAddr(BankLinks)                1A20

set ImageAddr(ScanResumeVFO)		1A40 ;# 0=pause, 1=busy ?
set ImageAddr(ScanResumeMem)		1A48 ;# 0=pause, 1=busy ?
set ImageAddr(ScanResumeDW)		1A50 ;# 0=pause, 1=busy ?

set ImageAddr(WhichDW)                	1A52 ;# values 0 - 8
set ImageAddr(SearchFreqFirst)		1B00
set ImageAddr(SearchModeFirst)		1B03
set ImageAddr(SearchFreqSecond)		1B20 ;# to 1B3C
set ImageAddr(SearchModeSecond)		1B23
set ImageAddr(DualWatchFirst)		1B40
set ImageAddr(DualWatchFirstMode)	1B43
set ImageAddr(DualWatchSecond)		1B60
set ImageAddr(DualWatchSecondMode)	1B63
set ImageAddr(BandStackFreq)		1B80 ;# 12 of them
set ImageAddr(BandStackMode)		1B83 ;# 12 of them
set ImageAddr(SmartSearchL)		1BB0
set ImageAddr(SmartSearchS)		1BB4
set ImageAddr(SmartSearchU)		1BB8
set ImageAddr(Priority)			1BBC
set ImageAddr(PriorityMode)		1BBF
set ImageAddr(PresetFreqs)		1BC0
set ImageAddr(PresetModes)		1BC3
set ImageAddr(PresetSteps)		1BC3
		# Note:
		# I don't know how the VR-120 uses WXFreq and WXMode.
		# I found 162.550 NFM located here in image
		# files of all VR-120s, even after a reset.
		# This may be related to Preprogrammed Frequency
		# Channel number 11 (see VR-120 manual pg. 17).
		#	 - Bob Parnass, August 12, 2002.
set ImageAddr(WXFreq)			1BEC
set ImageAddr(WXMode)			1BEF
set ImageAddr(SearchFreqFirstD)		1BF0 ;# Search band L--
set ImageAddr(SearchModeFirstD)		1BF3
set ImageAddr(SearchFreqSecondD)	1BF4
set ImageAddr(SearchModeSecondD)	1BF7
set ImageAddr(DualWatchFirstD)		1BF8 ;# Dual Watch A--
set ImageAddr(DualWatchFirstModeD)	1BFB 
set ImageAddr(DualWatchSecondD)		1BFC ;# Dual Watch b--
set ImageAddr(DualWatchSecondModeD)	1BFF
set ImageAddr(AutoBand)			1C00 ;# to 1C9C
set ImageAddr(AutoBandSM)		1C03
set ImageAddr(BandEdgeFreq)		1CA0 ;# see p10 of the manual
set ImageAddr(BandEdgeMode)		1CA3


##########################################################
# Translation tables
##########################################################

# Scan resume condition
set ScanResume(PAUSE)	0
set ScanResume(BUSY)	1
set ScanResume(HOLD)	2

set RScanResume(0)	PAUSE
set RScanResume(1)	BUSY
set RScanResume(2)	HOLD


# Auto Band mode/step byte

set AutoBandMode(01)	NFM,AUTO
set AutoBandMode(11)	NFM,5
set AutoBandMode(21)	NFM,6.25
set AutoBandMode(31)	NFM,9
set AutoBandMode(41)	NFM,10
set AutoBandMode(51)	NFM,12.5
set AutoBandMode(61)	NFM,15
set AutoBandMode(71)	NFM,20
set AutoBandMode(81)	NFM,25
set AutoBandMode(91)	NFM,30
set AutoBandMode(a1)	NFM,50
set AutoBandMode(b1)	NFM,100

set AutoBandMode(02)	WFM,AUTO
set AutoBandMode(12)	WFM,5
set AutoBandMode(22)	WFM,6.25
set AutoBandMode(32)	WFM,9
set AutoBandMode(42)	WFM,10
set AutoBandMode(52)	WFM,12.5
set AutoBandMode(62)	WFM,15
set AutoBandMode(72)	WFM,20
set AutoBandMode(82)	WFM,25
set AutoBandMode(92)	WFM,30
set AutoBandMode(a2)	WFM,50
set AutoBandMode(b2)	WFM,100

set AutoBandMode(03)	AM,AUTO
set AutoBandMode(13)	AM,5
set AutoBandMode(23)	AM,6.25
set AutoBandMode(33)	AM,9
set AutoBandMode(43)	AM,10
set AutoBandMode(53)	AM,12.5
set AutoBandMode(63)	AM,15
set AutoBandMode(73)	AM,20
set AutoBandMode(83)	AM,25
set AutoBandMode(93)	AM,30
set AutoBandMode(a3)	AM,50
set AutoBandMode(b3)	AM,100


# Auto Band mode/step byte

set RAutoBandMode(NFM,AUTO)	01
set RAutoBandMode(NFM,5)	11
set RAutoBandMode(NFM,6.25)	21
set RAutoBandMode(NFM,9)	31
set RAutoBandMode(NFM,10)	41
set RAutoBandMode(NFM,12.5)	51
set RAutoBandMode(NFM,15)	61
set RAutoBandMode(NFM,20)	71
set RAutoBandMode(NFM,25)	81
set RAutoBandMode(NFM,30)	91
set RAutoBandMode(NFM,50)	a1
set RAutoBandMode(NFM,100)	b1

set RAutoBandMode(WFM,AUTO)	02
set RAutoBandMode(WFM,5)	12
set RAutoBandMode(WFM,6.25)	22
set RAutoBandMode(WFM,9)	32
set RAutoBandMode(WFM,10)	42
set RAutoBandMode(WFM,12.5)	52
set RAutoBandMode(WFM,15)	62
set RAutoBandMode(WFM,20)	72
set RAutoBandMode(WFM,25)	82
set RAutoBandMode(WFM,30)	92
set RAutoBandMode(WFM,50)	a2
set RAutoBandMode(WFM,100)	b2

set RAutoBandMode(AM,AUTO)	03
set RAutoBandMode(AM,5)		13
set RAutoBandMode(AM,6.25)	23
set RAutoBandMode(AM,9)		33
set RAutoBandMode(AM,10)	43
set RAutoBandMode(AM,12.5)	53
set RAutoBandMode(AM,15)	63
set RAutoBandMode(AM,20)	73
set RAutoBandMode(AM,25)	83
set RAutoBandMode(AM,30)	93
set RAutoBandMode(AM,50)	a3
set RAutoBandMode(AM,100)	b3


# Lamp operation
# Note: RLamp value of 3 has same effect as 2, namely OFF
set RLamp(0)	AUTO
set RLamp(1)	ON
set RLamp(2)	OFF

# Frequency Counter width (MHz)
set RCounter(0)		5
set RCounter(1)		10
set RCounter(2)		50
set RCounter(3)		100

# Frequency Counter width (MHz)
# Note: I tried a counter value of 04 but the radio's counter
# function quit almost immediately - Bob Parnass, Aug 9, 2002.
set Counter(5)		00
set Counter(10)		01
set Counter(50)		02
set Counter(100)	03


# Fast Step size

set Fstep(10kHz)	0
set Fstep(100kHz)	1
set Fstep(1MHz)		2
set Fstep(10MHz)	3
set Fstep(100MHz)	4

set RFstep(0)	10kHz
set RFstep(1)	100kHz
set RFstep(2)	1MHz
set RFstep(3)	10MHz
set RFstep(4)	100MHz


# Note: I tried a value of "0c" for step and the radio
# acted irrationally. - Bob Parnass, Aug 24, 2002

set Step(AUTO)		"00"
set Step(5)		"01"
set Step(6.25)		"02"
set Step(9)		"03"
set Step(10)		"04"
set Step(12.5)		"05"
set Step(15)		"06"
set Step(20)		"07"
set Step(25)		"08"
set Step(30)		"09"
set Step(50)		"0a"
set Step(100)		"0b"

set RStep(00)		"AUTO"
set RStep(01)		"5"
set RStep(02)		"6.25"
set RStep(03)		"9"
set RStep(04)		"10"
set RStep(05)		"12.5"
set RStep(06)		"15"
set RStep(07)		"20"
set RStep(08)		"25"
set RStep(09)		"30"
set RStep(0a)		"50"
set RStep(0b)		"100"

set Mode(M0) "0"
set Mode(M7) "7"
set Mode(M8) "8"
set Mode(M9) "9"
set Mode(MA) "A"
set Mode(MB) "B"
set Mode(MC) "C"
set Mode(MD) "D"
set Mode(ME) "E"
set Mode(MF) "F"

set Mode(AUTO) "0"
set Mode(NFM) "1"
set Mode(WFM) "2"
set Mode(AM) "3"

set RMode(0) AUTO
set RMode(1) NFM
set RMode(2) WFM
set RMode(3) AM

set BatterySaver(OFF) 0
set BatterySaver(1) 1
set BatterySaver(3) 2
set BatterySaver(5) 3
set BatterySaver(7) 4
set BatterySaver(9) 5

set RBatterySaver(0) "OFF"
set RBatterySaver(1) "1"
set RBatterySaver(2) "3"
set RBatterySaver(3) "5"
set RBatterySaver(4) "7"
set RBatterySaver(5) "9"

##########################################################
# Open the serial port.
# Notes:
#	This procedure sets the global variable Sid.
#
# Returns:
#	"" -ok
#	exit if error
##########################################################

proc OpenDevice {} \
{
	global Pgm
	global GlobalParam
	global Sid
	global tcl_platform

	set msg ""


	set platform $tcl_platform(platform) 
	switch -glob $platform \
		{
		{unix} \
			{
			if [ catch { open $GlobalParam(Device) "r+"} \
				Sid] \
				{
				set msg "Error while trying to open "
				append msg "serial port "
				append msg $GlobalParam(Device)
				}
			}
		{macintosh} \
			{
			if [ catch { open $GlobalParam(Device) "r+"} \
				Sid] \
				{
				set msg "Error while trying to open "
				append msg "serial port "
				append msg $GlobalParam(Device)
				}
			}
		{windows} \
			{
			if [ catch { open $GlobalParam(Device) RDWR} \
				Sid] \
				{
				set msg "Error while trying to open "
				append msg "serial port "
				append msg $GlobalParam(Device)
				}
			}
		default \
			{
			set msg "$Pgm error: Platform $platform "
			append msg "not supported."
			}
		}

	waiter 500

	# Set up the serial port parameters (similar to stty)
	if { ($msg != "") || ([catch {fconfigure $Sid \
		-buffering full \
		-translation binary \
		-mode 9600,e,8,1 -blocking 1}] ) } \
		{
		set msg "$Pgm error: "
		set msg [append msg "Cannot configure serial port\n"]
		set msg [append msg "$GlobalParam(Device)"]
		}

	if {$msg != ""} \
		{
		Tattle $msg
		tk_dialog .opnerror "Serial port error" \
			$msg error 0 OK
		exit
		}
	waiter 1000


	return "" 
}
##########################################################
#
# Initialize a few global variables.
#
# Return the pathname to a configuration file in the user's
# HOME directory
#
# Returns:
#	list of 2 elements:
#		-name of configuration file
#		-name of label file
#
##########################################################
proc InitStuff { } \
{
	global argv0
	global DisplayFontSize
	global env
	global Home
	global Pgm
	global RootDir
	global tcl_platform


	set platform $tcl_platform(platform) 
	switch -glob $platform \
		{
		{unix} \
			{
			set Home $env(HOME)
			set rcfile [format "%s/.tk120rc" $Home]
			set labelfile [format "%s/.tk120la" $Home]

			set DisplayFontSize "Courier 56 bold"
			}
		{macintosh} \
			{

			# Configuration file should be
			# named $HOME/.tk120rc

			# Use forward slashes within Tcl/Tk
			# instead of colons.

			set Home $env(HOME)
			regsub -all {:} $Home "/" Home
			set rcfile [format "%s/.tk120rc" $Home]
			set labelfile [format "%s/.tk120la" $Home]

			# The following font line may need changing.
			set DisplayFontSize "Courier 56 bold"
			}
		{windows} \
			{

			# Configuration file should be
			# named $tk120/tk120.ini
			# Use forward slashes within Tcl/Tk
			# instead of backslashes.

			set Home $env(tk120)
			regsub -all {\\} $Home "/" Home
			set rcfile [format "%s/tk120.ini" $Home]
			set labelfile [format "%s/tk120.lab" $Home]

			set DisplayFontSize "Courier 28 bold"
			}
		default \
			{
			puts "Operating System $platform not supported."
			exit 1
			}
		}
	set Home $env(HOME)
	# set Pgm [string last "/" $argv0]


	set lst [list $rcfile $labelfile]
	return $lst
}

###################################################################
# Disable computer control of radio.
###################################################################
proc CloseDevice { } \
{
	global Sid

	after 500

	catch {close $Sid}
	return
}


###################################################################
#
# Send "command" to radio.
# Write command to error stream if Debug flag is set.
#
###################################################################
proc SendCmd { Sid cmd } \
{
	global GlobalParam


	if { $GlobalParam(Debug) > 0 } \
		{
		binary scan $cmd "H*" s

		# Insert a space between each pair of hex digits
		# to improve readability.

		regsub -all ".." $s { \0} s
		set msg ""
		set msg [ append msg "---> " $s]
		Tattle $msg
		}

	# Write data to serial port.

	puts -nonewline $Sid $cmd
	flush $Sid

	if { $GlobalParam(Debug) > 0 } \
		{
		# puts stderr "SendCmd: wrote data to radio."
		}
	return
}



###################################################################
# Send a msg to radio in preparation for writing an image
# to the radio.  Read back msg to clear the IO stream.
#
# String written will be whatever is stored in
# the variable GlobalParam(PreambleRead) (with spaces are
# removed):
#
#	0c 01 41 31 38 12 00 c9 (USA models)
#	0c 01 41 31 38 11 01 c9 (Export model)
#	0c 01 41 31 38 10 01 c8 (Japanese model)
#	
#
# Note: The last byte is a checksum calculated over
#	the previous 7 bytes.
#
#	Inputs: none
#		
###################################################################
proc WritePreamble { } \
{
	global GlobalParam
	global Sid

	# Delete all spaces from preamble before writing it.
	regsub -all { } $GlobalParam(PreambleRead) "" p

	set cmd ""
	append cmd [binary format "H*" $p ]

#	if {$GlobalParam(WhichModel) == "USA"} \
#		{
#		# append cmd [binary format "H2H2H2H2H2H2H2H2" \
#			# 0C 01 41 31 38 12 00 C9 ]
#		append cmd [binary format "H*" \
#			$GlobalParam(PreambleRead) ]
#		} \
#	elseif {$GlobalParam(WhichModel) == "Japanese"} \
#		{
#		append cmd [binary format "H2H2H2H2H2H2H2H2" \
#			0c 01 41 31 38 10 01 c8 ]
#		} \
#	else \
#		{
#		# Export and other models
#		append cmd [binary format "H2H2H2H2H2H2H2H2" \
#			0c 01 41 31 38 11 01 c9 ]
#		}
#

	SendCmd $Sid $cmd


	# Read back the same msg.
	ReadRx 8

	return
}


###################################################################
# Send a msg to radio after writing an image
# to the radio.  Read back msg to clear the IO stream.
#
# String written will be:
#
#	0c 02 41 31 38 00 00 b8
#
# Note: The last byte is a checksum calculated over
#	the previous 7 bytes.
#
#	Inputs: none
#		
###################################################################
proc WritePostamble { } \
{
	global GlobalParam
	global Sid

	set cmd ""

	append cmd [binary format "H2H2H2H2H2H2H2H2" \
		0c 02 41 31 38 00 00 b8 ]


	SendCmd $Sid $cmd


	# Read back the same msg.
	ReadRx 8

	return
}

###################################################################
# Send positive ack to radio.
#	Inputs: none
#		
#
###################################################################
proc SendAck { } \
{
	global Sid

	set cmd ""
	set cmd [ append cmd [binary format "H2" 03 ]]
	set cmd [ append cmd [binary format "H2" 06 ]]
	set cmd [ append cmd [binary format "H2" 09 ]]

	SendCmd $Sid $cmd


	# Read the ok ack.
	ReadRx 3

	return
}


###################################################################
# Read acknowledgement from the radio.
#
#	Inputs: none
#	Returns:
#		0	-ok
#		1	-error
#		
# The string read should be:  0C 06 00
#
###################################################################
proc ReadAck { } \
{
	global Sid

	# Read the acknowledgement from the radio.

	# puts stderr "ReadAck: going to read acknowledgement from radio."
	set line [ReadRx 3]
	# puts stderr "ReadAck: read acknowledgement from radio."

	set c0 [string range $line 0 0]
	binary scan $c0 "H*" c0

	set c1 [string range $line 1 1]
	binary scan $c1 "H*" c1

	set c2 [string range $line 2 2]
	binary scan $c2 "H*" c2

	# If this is a positive ack.

	if { ($c0 == "0c") \
		&& ($c1 == "06")
		&& ($c2 == "00") } \
		{
		set code 0
		} \
	else \
		{
		puts stderr "Received bad acknowledgement from radio."
		set code 1
		}

	return $code

}


###################################################################
# Read a 32 byte msg from the serial port.
#
# Returns:
#		The message unless there was an error.
###################################################################
proc ReadMsg { } \
{
	global GlobalParam
	global Pgm
	global Sid


	set buf ""

	set buf1 [ReadRx 8]

	set code [string range $buf1 1 1]
	binary scan $code "H2" hcode

	if { $hcode != "03" } \
		{
		# Radio probably sent a postamble message,
		# which means there is no more data to read.


		binary scan $buf1 "H*" x

		regsub -all ".." $x { \0} x

		set GlobalParam(PostambleRead) $x
		return ""
		}
	set buf2 [ReadRx 30]

	append buf $buf1 $buf2

	# The last byte of each message is a checksum.

	set dbuf [string range $buf 0 36] 
	set cksum [string range $buf 37 37] 
	set ccksum [CalcCheckSum $dbuf]

	binary scan $cksum "H*" icksum
	scan $icksum "%x" cksum

#	puts stderr [format "CHECKSUM radio: %s, calculated: %s\n" \
#			$cksum $ccksum]

	if {$cksum != $ccksum} \
		{
		set msg [format \
			"%s: error, checksum mismatch, radio: %s, calculated: %s\n" \
			$Pgm $cksum $ccksum]
		Tattle $msg
		tk_dialog .error "Checksum error while reading" \
			$msg error 0 OK
		exit
		}

	# Strip off the first 5 bytes, which are:
	# two address bytes,
	# three sequence number bytes
	
	set line [string range $buf 5 36] 


	return $line
}

###################################################################
# Read 'n' bytes from the serial port.
#
# Returns:
#		The message unless there was an error.
###################################################################
proc ReadRx { n } \
{
	global GlobalParam
	global Sid


	set line ""

	if { $GlobalParam(Debug) > 0 } \
		{
		# puts stderr "ReadRx: entered. Going to read $n bytes."
		}

	for {set i 0} {($i < $n) && (![eof $Sid])} {incr i} \
		{
		set b [read $Sid 1]
		# puts stderr "ReadRx: got 1 char"
		set line [append line $b]
		}

	if { $GlobalParam(Debug) > 0 } \
		{
		set msg "<--- "
		binary scan $line "H*" x

		regsub -all ".." $x { \0} x

		set msg [append msg $x]
		Tattle $msg
		}

	return $line
}


##########################################################
# Copy memory image to radio
#
# Returns:
#	0	-ok
#	1	-error
##########################################################
proc WriteImage { }\
{
	global GlobalParam
	global Mimage
	global Nmessages
	global Sid

	# Open serial port.
	OpenDevice

	set totmsgs [expr {$Nmessages + $GlobalParam(Wadjust)}]

	# Create and display progress bar.
	toplevel .pbw
	wm title .pbw "Writing to VR-120"
	grab set .pbw
	set p [MakeWaitWindow .pbw.g 0 PaleGreen]
	set pc 0
	gauge_value $p $pc
	update


	set db 0


	WritePreamble
	if {[ReadAck]} \
		{

		# Zap the progress bar.
		grab release .pbw
		catch {destroy .pbw}

		# Close serial port
		CloseDevice

		return 1
		}	

	set bptr 0

	# For each message.
	for {set i 0} {$i < $Nmessages} {incr i} \
		{

		set line ""
		append line [binary format "H2" 03 ]
		append line [binary format "H2" 03 ]

		# A 3 byte message counter
		append line [binary format "H2" 00 ]
		append line [binary format "S" $i ]

		# Copy the next chunk of the image, but skip
		# the checksum byte.

		set end [expr {$bptr + 31}]
		set s [string range $Mimage $bptr $end]
		append line $s

		#
		# Calulate and append the checksum byte
		#

		# Checksum is decimal.
		set cksum [CalcCheckSum $line]

		# Convert checksum to hexadecimal.
		set hcksum [format "%02x" $cksum]

		set bcksum [binary format "H2" $hcksum ]


		append line $bcksum
		SendCmd $Sid $line
		ReadRx 38
		if {[ReadAck]} \
			{
			# Zap the progress bar.
			grab release .pbw
			catch {destroy .pbw}

			# Close serial port
			CloseDevice

			return 1
			}	



		incr bptr 32

		# Update progress bar widget.
		set pc [ expr $i * 100 / $totmsgs ]
		if {$pc >= 100.0} \
			{
			set pc 99
			}
		gauge_value $p $pc
		}

	set ffmsg [Padff 32]
	set recno $i


	# Write extra messages to radio because it requires
	# them.  The messages contain 32 \xff bytes, checksum.

	for {set i 0} {$i < $GlobalParam(Wadjust)} {incr i} \
		{
		set line ""
		append line [binary format "H2" 03 ]
		append line [binary format "H2" 03 ]

		# A 3 byte message counter
		append line [binary format "H2" 00 ]
		append line [binary format "S" $recno ]

		append line $ffmsg

		# Calulate and append the checksum byte

		# Checksum is decimal.
		set cksum [CalcCheckSum $line]

		# Convert checksum to hexadecimal.
		set hcksum [format "%02x" $cksum]

		set bcksum [binary format "H2" $hcksum ]


		append line $bcksum
		SendCmd $Sid $line
		ReadRx 38
		if {[ReadAck]} \
			{
			# Zap the progress bar.
			grab release .pbw
			catch {destroy .pbw}

			# Close serial port
			CloseDevice

			return 1
			}	


		# Update progress bar widget.
		set pc [ expr $recno * 100 / $totmsgs ]
		if {$pc >= 100.0} \
			{
			set pc 99
			}
		gauge_value $p $pc

		incr recno
		}

	WritePostamble

	if {[ReadAck]} \
		{

		# Zap the progress bar.
		grab release .pbw
		catch {destroy .pbw}

		# Close serial port
		CloseDevice

		return 1
		}


	# Zap the progress bar.
	grab release .pbw
	catch {destroy .pbw}

	# Close serial port
	CloseDevice

	return 0
}


##########################################################
# Copy memory image from radio
##########################################################
proc ReadImage { } \
{
	global GlobalParam
	global Mimage
	global Nmessages



	# Create and display progress bar.
	toplevel .pbw
	wm title .pbw "Reading VR-120"
	grab set .pbw
	set p [MakeWaitWindow .pbw.g 0 PaleGreen]
	set pc 0
	gauge_value $p $pc
	update


	set Mimage ""

	# Open and configure the serial port.

	OpenDevice
	set line [ReadRx 8]

	binary scan $line "H*" x
	regsub -all ".." $x { \0} x

	set GlobalParam(PreambleRead) $x

	SendAck

	# For each message.
	for {set i 0} {$i < 500} {incr i} \
		{
		set line [ReadMsg]
		append Mimage $line
		SendAck

		# Update progress bar widget.
		set pc [ expr {$i * 100 / $Nmessages} ]
		if {$pc >= 100.0} \
			{
			set pc 99
			}
		gauge_value $p $pc

		if {[string length $line] == 0} \
			{
			# No more data to read.
			break
			}
		}

	set GlobalParam(NmsgsRead) $i

	# Note: The last several messages which don't contain
	# useful info.



	# Zap the progress bar.
	grab release .pbw
	destroy .pbw

	# Close serial port.
	CloseDevice

	return 0
}



###################################################################
# this takes a string and converts the
# first character in it to an integer
#   in the range 0-255
#
# if the string is empty, returns an empty string
###################################################################

proc Char2Int { c } \
{
	set tmp ""

	set n [binary scan $c "c" tmp]

	if { ($n == 1) && ($tmp < 0) } \
		{
		# Force negative number to be positive

		set tmp [expr $tmp + 256]
		}
	return "$tmp"
}


###################################################################
# Calculate the modulo 256 checksum byte for a string by
# summing all the ascii character values, modulo 256.
###################################################################

proc CalcCheckSum { s } \
{

	set len [string length $s]
	set sum 0

	for {set i 0} {$i < $len} {incr i} \
		{
         	set tmp [Char2Int [string index $s $i]]
         	set sum [expr {$sum + $tmp}]
		}

	set sum [expr $sum % 256]
	return $sum
 }

###################################################################
# Create a string of "n" bytes where each byte is \xff (255 decimal).
###################################################################

proc Padff { n } \
{
	set ffrecd ""
	set byte [binary format "H2" ff]

	for {set i 0} {$i < $n} {incr i} \
		{
		append ffrecd $byte
		}
	return $ffrecd
}
