###################################################################
# This file is part of tk120, a utility program for the
# Yaesu VR-120 receiver.
# 
#    Copyright (C) 2001 - 2005, Bob Parnass
#					AJ9S
# 
# tk120 is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; either version 2 of the License,
# or (at your option) any later version.
# 
# tk120 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with tk120; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# 02111-1307  USA
###################################################################


proc MakeGui { } \
{
	global Cht
	global Chvector
	global GlobalParam
	global ReadRadioFlag
	global TemplateSavedFlag

	set Cht ""

	set GlobalParam(TemplateFilename) untitled.tk1

	set TemplateSavedFlag no
	set ReadRadioFlag no

	# Set custom font and colors.

	SetAppearance

	
	###############################################################
	# Menu bar along the top edge.
	###############################################################
	set fr_menubar [MakeMenuBar .mb]
	set mf [frame .mainframe]
	set fr_line1 [frame $mf.line1]
	set fr_lim [frame $mf.lim]
	
	set fr_misc [MakeMiscFrame $fr_line1.omsg]
	set fr_display [MakeDisplayFrame $fr_line1.dis]
	set fr_tweaks [MakeTweaksFrame $fr_line1.tw]
	set fr_ss [MakeSmartSearchFrame $fr_line1.ss]
	set fr_title [MakeTitleFrame $fr_line1.title]
	pack $fr_title -side right -fill y
	pack $fr_misc $fr_display $fr_ss -side left -fill y

	if { $GlobalParam(ExperimentalTweaks) == "on" } \
		{
		pack $fr_tweaks -side left -fill y
		}

	###############################################################
	# VFO controls window
	###############################################################
	toplevel .vfo
	set fr_vfo .vfo.ctls
	frame $fr_vfo -relief groove

	set fr_search [MakeSearchFrame $fr_vfo.search]
	set fr_skip [MakeSkipFrame $fr_vfo.skip]
	set fr_autoband [MakeAutoBandFrame $fr_vfo.autoband]

	pack $fr_search $fr_skip $fr_autoband \
		-side left \
		-fill both -expand true

	pack $fr_vfo
	
	# Prevent user from closing the VFO controls window unless
	# he elects to exit the entire program.
	wm protocol .vfo WM_DELETE_WINDOW {ExitApplication}
	wm title .vfo "tk120 VFO Settings"
	

	###############################################################
	# Memory Bank controls window
	###############################################################

	toplevel .mbank
	set mbank .mbank.ctls
	frame $mbank -relief groove
	wm title .mbank "tk120 Memory Bank"

	set fr_bank [MakeMemoryBankFrame $mbank.bank]

	pack $fr_bank -side left -fill both -expand true
	pack $mbank -side left -fill both -expand true
	
	# Prevent user from closing the Bank window unless
	# he elects to exit the entire program.
	wm protocol .mbank WM_DELETE_WINDOW {ExitApplication}

	###############################################################
	# Band controls window
	###############################################################
	toplevel .band
	set fr_band .band.ctls
	frame $fr_band -relief groove

	set fr_bandedge [MakeBandEdgeFrame $fr_band.bandedge]
	set fr_bandstack [MakeBandStackFrame $fr_band.bandstack]

	pack $fr_bandedge \
		$fr_bandstack \
		-side left \
		-fill both -expand true

	pack $fr_band
	
	# Prevent user from closing the Band controls window unless
	# he elects to exit the entire program.
	wm protocol .band WM_DELETE_WINDOW {ExitApplication}
	wm title .band "tk120 Band Stacking Settings"
	

	###############################################################
	# Secondary controls window
	###############################################################
	toplevel .controls
	set ctls .controls.ctls
	frame $ctls -relief groove

	set fr_preset [MakePresetFrame $ctls.preset]
	# set fr_autoband [MakeAutoBandFrame $ctls.autoband]
	set fr_dualwatch [MakeDWFrame $ctls.dw]
	set fr_com [MakeCommFrame $ctls.com]

	pack $fr_preset $fr_dualwatch \
		-side left \
		-fill both -expand true
	
	

	pack $fr_com -side left -fill both -expand true

	
	
	
	set Chvector ""
	
	pack $fr_menubar -side top -fill x -pady 3 -padx 3
	pack $fr_line1 -side top -fill x -pady 3 -padx 3
	pack $fr_lim -side top -fill x -pady 3 -padx 3
	
	
	pack $ctls -side top -fill both -expand true -padx 3 -pady 3
	pack .mainframe -side top -fill both -expand true
	
	update idletasks
	
	###############################################################
	#  Ask the window manager to catch the delete window
	#  event.
	###############################################################
	wm protocol . WM_DELETE_WINDOW {ExitApplication}
	
	# Prevent user from shrinking or expanding main window.
	wm minsize . [winfo width .] [winfo height .]
	wm maxsize . [winfo width .] [winfo height .]
	
	wm protocol .controls WM_DELETE_WINDOW {ExitApplication}
	wm title .controls "tk120 Secondary Controls"
	
	
	# Prevent user from overshrinking or expanding controls window.
	wm minsize .controls [winfo width .controls] [winfo height .controls]
	wm maxsize .controls [winfo width .controls] [winfo height .controls]
	
	
	# Prevent user from shrinking or expanding window.
	wm minsize .vfo [winfo width .vfo] [winfo height .vfo]
	wm maxsize .vfo [winfo width .vfo] [winfo height .vfo]

	# Prevent user from shrinking or expanding window.
	wm minsize .band [winfo width .band] [winfo height .band]
	wm maxsize .band [winfo width .band] [winfo height .band]


	# Force main window to appear on top by hiding, then
	# then showing it.
	wm withdraw .
	wm deiconify .

	return
}


###################################################################
# Alter color and font appearance based on user preferences.
###################################################################
proc SetAppearance { } \
{
	global GlobalParam

	if {$GlobalParam(Font) != "" } \
		{
		# Designate a custom font for most widgets.
		option add *font $GlobalParam(Font)
		}

	if {$GlobalParam(BackGroundColor) != "" } \
		{
		# Designate a custom background color for most widgets.
		option add *background $GlobalParam(BackGroundColor)
		}

	if {$GlobalParam(ForeGroundColor) != "" } \
		{
		# Designate a custom foreground color for most widgets.
		option add *foreground $GlobalParam(ForeGroundColor)
		}

	if {$GlobalParam(TroughColor) != "" } \
		{
		# Designate a custom slider trough color
		# for most scale widgets.
		option add *troughColor $GlobalParam(TroughColor)
		}

	return
}



##########################################################
# Check if the configuration file exists.
# If it exits, return 1.
#
# Otherwise, prompt the user to select the
# serial port.
##########################################################

proc FirstTimeCheck { Rcfile } \
{
	global AboutMsg
	global GlobalParam
	global Libdir
	global tcl_platform

	if { [file readable $Rcfile] == 1 } \
		{
		return 0
		}

	tk_dialog .about "About tk120" \
		$AboutMsg info 0 OK

	# No readable config file found.
	# Treat this as the first time the user has run the program.

	# Create a new window with radio buttions and
	# an entry field so user can designate the proper
	# serial port.

	set msg "Please identify the serial port to which\n"
	set msg [append msg "your VR-120 receiver is connected."]

	toplevel .serialport
	set sp .serialport

	label $sp.intro -text $msg

	frame $sp.rbframe
	set fr $sp.rbframe

	if { $tcl_platform(platform) == "windows" } \
		{
		# For Windows.
		radiobutton $fr.com1 -text COM1: -variable port \
			-value {COM1:}
		radiobutton $fr.com2 -text COM2: -variable port \
			-value {COM2:} 
		radiobutton $fr.com3 -text COM3: -variable port \
			-value {COM3:} 
		radiobutton $fr.com4 -text COM4: -variable port \
			-value {COM4:} 

		pack $fr.com1 $fr.com2 $fr.com3 $fr.com4 \
			-side top -padx 3 -pady 3 -anchor w

		} \
	else \
		{
		# For unix, mac, etc..
		radiobutton $fr.s0 -text /dev/ttyS0 -variable port \
			-value {/dev/ttyS0} 
		radiobutton $fr.s1 -text /dev/ttyS1 -variable port \
			-value {/dev/ttyS1} 
		radiobutton $fr.s2 -text /dev/ttyS2 -variable port \
			-value {/dev/ttyS2} 
		radiobutton $fr.s3 -text /dev/ttyS3 -variable port \
			-value {/dev/ttyS3} 
		radiobutton $fr.s4 -text /dev/ttyS4 -variable port \
			-value {/dev/ttyS4} 
		radiobutton $fr.s5 -text /dev/ttyS5 -variable port \
			-value {/dev/ttyS5} 

		pack \
			$fr.s0 $fr.s1 $fr.s2 \
			$fr.s3 $fr.s4 $fr.s5 \
			-side top -padx 3 -pady 3 -anchor w

		}

	radiobutton $fr.other -text "other (enter below)" \
		-variable port \
		-value other

	entry $fr.ent -width 30 -textvariable otherport

	pack $fr.other $fr.ent \
		-side top -padx 3 -pady 3 -anchor w

	button $sp.ok -text "OK" \
		-command \
			{ \
			global GlobalParam

			if {$port == "other"} \
				{
				set GlobalParam(Device) $otherport
				} \
			else \
				{
				set GlobalParam(Device) $port
				}
			# puts stderr "entered $GlobalParam(Device)"
			}

	button $sp.exit -text "Exit" \
		-command { exit }

	pack $sp.intro -side top -padx 3 -pady 3
	pack $fr -side top -padx 3 -pady 3
	pack $sp.ok $sp.exit -side left -padx 3 -pady 3 -expand true



	bind $fr.ent <Key-Return> \
		{
		global GlobalParam
		set GlobalParam(Device) $otherport
		}

	wm title $sp "Select serial port"
	wm protocol $sp WM_DELETE_WINDOW {exit}

	set errorflag true

	while { $errorflag == "true" } \
		{
		tkwait variable GlobalParam(Device)

		if { $tcl_platform(platform) != "unix" } \
			{
			set errorflag false
			break
			}

		# The following tests do not work properly
		# in Windows. That is why we won't perform
		# the serial port tests when running Windows version.

		if { ([file readable $GlobalParam(Device)] != 1) \
			|| ([file writable $GlobalParam(Device)] != 1)}\
			{
			# Device must be readable, writable

			bell
			tk_dialog .badport "Serial port problem" \
				"Serial port problem" error 0 OK
			} \
		else \
			{
			set errorflag false
			}
		}

	destroy $sp
	return 1
}

##########################################################
# ExitApplication
#
# This procedure can do any cleanup necessary before
# exiting the program.
#
# Disable computer control of the radio, then quit.
##########################################################
proc ExitApplication { } \
{
	global Lid
	global Lfilename
	global TemplateSavedFlag
	global ReadRadioFlag

	if { ($ReadRadioFlag == "yes") \
		&&  ($TemplateSavedFlag == "no") } \
		{
		set msg "You did not save the template data"
		append msg " in a file."

		set result [tk_dialog .sav "Warning" \
			$msg \
			warning 0 Cancel Exit ]

		if {$result == 0} \
			{
			return
			}
		}

	SaveSetup
	# CloseDevice

	if { [info exists Lfilename] } \
		{
		if { $Lfilename != "" } \
			{
			# Close log file.
			close $Lid
			}
		}
	exit
}


##########################################################
# NoExitApplication
#
# This procedure prevents the user from
# killing the window.
##########################################################
proc NoExitApplication { } \
{

	set response [tk_dialog .quitit "Exit?" \
		"Do not close this window." \
		warning 0 OK ]

	return
}
##########################################################
#
# Scroll_Set manages optional scrollbars.
#
# From "Practical Programming in Tcl and Tk,"
# second edition, by Brent B. Welch.
# Example 27-2
#
##########################################################

proc Scroll_Set {scrollbar geoCmd offset size} {
	if {$offset != 0.0 || $size != 1.0} {
		eval $geoCmd;# Make sure it is visible
		$scrollbar set $offset $size
	} else {
		set manager [lindex $geoCmd 0]
		$manager forget $scrollbar								;# hide it
	}
}


##########################################################
#
# Listbox with optional scrollbars.
#
#
# Inputs: basename of configuration file
#
# From "Practical Programming in Tcl and Tk,"
# second edition, by Brent B. Welch.
# Example 27-3
#
##########################################################

proc Scrolled_Listbox { f args } {
	frame $f
	listbox $f.list \
		-font {courier 12} \
		-xscrollcommand [list Scroll_Set $f.xscroll \
			[list grid $f.xscroll -row 1 -column 0 -sticky we]] \
		-yscrollcommand [list Scroll_Set $f.yscroll \
			[list grid $f.yscroll -row 0 -column 1 -sticky ns]]
	eval {$f.list configure} $args
	scrollbar $f.xscroll -orient horizontal \
		-command [list $f.list xview]
	scrollbar $f.yscroll -orient vertical \
		-command [list $f.list yview]
	grid $f.list $f.yscroll -sticky news
	grid $f.xscroll -sticky news

	grid rowconfigure $f 0 -weight 1
	grid columnconfigure $f 0 -weight 1

	return $f.list
}

#
###########################################################
##
## Create a scrollable frame.
##
##
## From "Effective Tcl/Tk Programming,"
## by Mark Harrison and Michael McLennan.
## Page 121.
##
###########################################################
#
#proc ScrollformCreate { win } \
#{
#
#	frame $win -class Scrollform -relief groove -borderwidth 3
#	scrollbar $win.sbar -command "$win.vport yview"
#	pack $win.sbar -side right -fill y
#
#	canvas $win.vport -yscrollcommand "$win.sbar set"
#	pack $win.vport -side left -fill both -expand true
#
#	frame $win.vport.form
#	$win.vport create window 0 0 -anchor nw \
#		-window $win.vport.form
#
#	bind $win.vport.form <Configure> "ScrollFormResize $win"
#	return $win
#}
#
#proc ScrollFormResize { win } \
#{
#	set bbox [ $win.vport bbox all ]
#	set wid [ winfo width $win.vport.form ]
#	$win.vport configure -width $wid \
#		-scrollregion $bbox -yscrollincrement 0.1i
#}
#
#proc ScrollFormInterior { win } \
#{
#	return "$win.vport.form"
#}

##########################################################
#
# Return the bank selected from the channel selector listbox
#
##########################################################

proc BankSelected { w } \
{
	set line [ ListSelected $w ]
	set line [string trimleft $line " "]
	regsub " .*" $line "" ch
	return "$ch"
}

##########################################################
#
# Return the channel selected from the channel selector listbox
#
##########################################################

proc ChSelected { w } \
{
	set line [ ListSelected $w ]
	set line [string trimleft $line " "]

	regsub {^[0-9]} $line "" line
	regsub {^[0-9]} $line "" line
	set line [string trimleft $line " "]


	regsub " .*" $line "" ch
	return "$ch"
}


##########################################################
# Contruct the top row of pulldown menus
##########################################################
proc MakeMenuBar { f } \
{
	global AboutMsg
	global Device
	global FileTypes
	global GlobalParam
	global Pgm
	global Version

	# File pull down menu
	frame $f -relief groove -borderwidth 3

	menubutton $f.file -text "File" -menu $f.file.m \
		-underline 0
	menubutton $f.view -text "View" -menu $f.view.m \
		-underline 0
	menubutton $f.data -text "Data" -menu $f.data.m \
		-underline 0
	menubutton $f.radio -text "Radio" -menu $f.radio.m \
		-underline 0
	menubutton $f.help -text "Help" -menu $f.help.m \
		-underline 0
	
	
	menu $f.view.m
	AddView $f.view.m

	menu $f.data.m
	AddData $f.data.m
	
	menu $f.help.m


	$f.help.m add command -label "Readme" \
		-underline 0 \
		-command { \
			set helpfile [format "%s/README" $Libdir ]
			set win [textdisplay_create "README"]
			textdisplay_file $win $helpfile
			}
	
	$f.help.m add command -label "Tcl info" \
		-underline 0 \
		-command { \
			tk_dialog .about "Tcl info" \
				[HelpTclInfo] info 0 OK
			}

	$f.help.m add command -label "License" \
		-underline 0 \
		-command { \
			set helpfile [format "%s/COPYING" $Libdir ]
			set win [textdisplay_create "Notice"]
			textdisplay_file $win $helpfile
			}
	
	$f.help.m add command -label "About tk120" \
		-underline 0 \
		-command { \
			tk_dialog .about "About tk120" \
				$AboutMsg info 0 OK
			}
	
	menu $f.file.m -tearoff no

	$f.file.m add command -label "Open ..." \
		-underline 0 \
		-command {OpenTemplate .mainframe}
	
	$f.file.m add command -label "Save" \
		-underline 0 \
		-command {SaveTemplate .mainframe 0}

	$f.file.m add command -label "Save As ..." \
		-underline 0 \
		-command {SaveTemplate .mainframe 1}

	$f.file.m add separator

	$f.file.m add command -label "Import memory channels ..." \
		-underline 0 \
		-command {\
			ImportChannels .
			}

	$f.file.m add command -label "Export memory channels ..." \
		-underline 0 \
		-command {ExportChannels .mainframe}

	$f.file.m add separator

	$f.file.m add command -label "Exit" \
		-underline 1 \
		-command { ExitApplication}
	
	menu $f.radio.m -tearoff no
	AddRadio $f.radio.m


	pack $f.file $f.view $f.data $f.radio -side left -padx 10
	pack $f.help -side right
	
	update
	return $f
}



proc MakeScrollPane {w x y} {
   frame $w -class ScrollPane -width $x -height $y
   canvas $w.c -xscrollcommand [list $w.x set] -yscrollcommand [list $w.y set]
   scrollbar $w.x -orient horizontal -command [list $w.c xview]
   scrollbar $w.y -orient vertical   -command [list $w.c yview]
   set f [frame $w.c.content -borderwidth 0 -highlightthickness 0]
   $w.c create window 0 0 -anchor nw -window $f
   grid $w.c $w.y -sticky nsew
   grid $w.x      -sticky nsew
   grid rowconfigure    $w 0 -weight 1
   grid columnconfigure $w 0 -weight 1
   # This binding makes the scroll-region of the canvas behave correctly as
   # you place more things in the content frame.
   bind $f <Configure> [list Scrollpane_cfg $w %w %h]
   $w.c configure -borderwidth 0 -highlightthickness 0
   return $f
}
proc Scrollpane_cfg {w wide high} {
   set newSR [list 0 0 $wide $high]
	return
   if {![string equals [$w cget -scrollregion] $newSR]} {
      $w configure -scrollregion $newSR
   }
}


###################################################################
# Check to see if someone is running another
# copy of  this program.
#
# Code Snippit by DJ Eaton.
# Ben Mesander helped with Mac OS X compatibility.
#
# Warning: works in Linux and Mac OS X only, not Solaris
#
###################################################################
proc CheckForDup { } \
{
	global argv0
	global Pgm

	set filename [lindex [split $argv0 "/"] end]
	set ppid [lindex \
		[exec ps xwww \| grep -i wish \| grep $filename] 0]

	if {$ppid != [pid] } \
		{
		puts stderr "$Pgm: A copy of this program is already running.\n"
		exit 1
		}
	unset ppid filename
	return
}


##########################################################
# Add widgets to the view menu
##########################################################
proc AddView { m } \
{
	global GlobalParam


	# Change font.

	if {$GlobalParam(Font) == ""} \
		{
		set msg "Change Font"
		} \
	else \
		{
		set msg [format "Change Font (%s)" $GlobalParam(Font)]
		}

	$m add command -label $msg -command \
		{
		set ft [font_select]
		if {$ft != ""} \
			{
			set GlobalParam(Font) $ft

			set msg "The change will take effect next "
			set msg [append msg "time you start tk120."]

			tk_dialog .wcf "Change Appearance" \
				$msg info 0 OK
			}
		}

	$m add command -label "Restore Original Font" -command \
		{
		set GlobalParam(Font) ""
		set msg "The change will take effect next "
		set msg [append msg "time you start tk120."]

		tk_dialog .wcf "Change Appearance" $msg info 0 OK
		}

	$m add separator

	$m add command -label "Change Panel Color" -command \
		{
		set col [tk_chooseColor -initialcolor #d9d9d9]
		if {$col != ""} \
			{
			set GlobalParam(BackGroundColor) $col

			set msg "The change will take effect next "
			set msg [append msg "time you start tk120."]

			tk_dialog .wcf "Change Appearance" \
				$msg info 0 OK
			}
		}

	$m add command -label "Change Lettering Color" -command \
		{
		set col [tk_chooseColor -initialcolor black]
		if {$col != ""} \
			{
			set GlobalParam(ForeGroundColor) $col

			set msg "The change will take effect next "
			set msg [append msg "time you start tk120."]

			tk_dialog .wcf "Change Appearance" \
				$msg info 0 OK
			}
		}

	$m add command -label "Change Slider Trough Color" -command \
		{
		set col [tk_chooseColor -initialcolor #c3c3c3]
		if {$col != ""} \
			{
			set GlobalParam(TroughColor) $col

			set msg "The change will take effect next "
			set msg [append msg "time you start tk120."]

			tk_dialog .wcf "Change Appearance" \
				$msg info 0 OK
			}
		}

	$m add separator


	# Helpful tips balloons


	$m add  checkbutton \
		-label "Balloon Help Windows" \
                -variable GlobalParam(BalloonHelpWindows) \
                -onvalue on \
                -offvalue off 

	$m add  checkbutton \
		-label "Experimental Tweaks" \
                -variable GlobalParam(ExperimentalTweaks) \
                -onvalue on \
                -offvalue off \
		-command { \
			set msg "The change will take effect next "
			set msg [append msg "time you start tk120."]
			tk_dialog .wcf "Change Appearance" \
				$msg info 0 OK
			}

	return
}


##########################################################
# Add widgets to the Data menu
##########################################################
proc AddData { m } \
{
	global GlobalParam


	set hint ""
	append hint "The Encode Image operation "
	append hint "is designed for use when "
	append hint "testing tk120."
	balloonhelp_for $m $hint

	$m add command -label "Validate data" \
		-command \
			{
			if {[ValidateData] == 0} \
				{
				tk_dialog .info "Valiate data" \
				"The data is ok." info 0 OK
				}
			}



	$m add command -label "Check for duplicate frequencies" \
		-command { CkDuplicate }

	$m add command -label "Encode Image" \
		-command { \
			if {[ValidateData] == 0} \
				{
				MakeWait
				EncodeImage
				KillWait
				}
			}

	$m add separator
	
	
	$m add command -label "Swap Memory Banks ..." \
		-command { MakeSwapFrame }

	$m add command -label "Sort Memory Channels ..." \
		-command { MakeSortFrame }


	$m add command -label "Clear All Memory Channels ..." \
		-command { ClearAllChannels }

	return
}



##########################################################
# Add choices to the Radio menu
##########################################################
proc AddRadio { m } \
{
	global GlobalParam
	global Libdir


	$m add command -label "Read from radio ..." \
		-command { \
			Radio2Template .mainframe
			update
			}
	
	$m add command -label "Write to radio ..." \
		-command { \
			Image2Radio .mainframe
			update
			}
	
	
	$m add separator

	# Note:
	# Gummed label on VR-120 box carries these
	# designations:
	#
	# DST:  EAN/JAN:
	# ---   ---------
	# JPN   AH005M001
	# USA   AH005M002
	# EXP   AH005M004

	$m add radiobutton \
		-label "USA model (DST: USA, EAN/JAN: AH005M002)" \
		-variable GlobalParam(WhichModel) \
		-value USA

	$m add radiobutton \
		-label "Japanese, Export, or other model" \
		-variable GlobalParam(WhichModel) \
		-value Other


	$m add separator


	$m add  checkbutton \
		-label "Debug" \
                -variable GlobalParam(Debug) \
                -onvalue "1" \
                -offvalue "0"

	return $m
}


##########################################################
#
# Create a progress gauge widget.
#
#
# From "Effective Tcl/Tk Programming,"
# by Mark Harrison and Michael McLennan.
# Page 125.
#
##########################################################
proc gauge_create {win {color ""}} \
{
	frame $win -class Gauge

	# set len [option get $win length Length]
	set len 300

	canvas $win.display -borderwidth 0 -background white \
		-highlightthickness 0 -width $len -height 20
	pack $win.display -expand yes -padx 10
	if {$color == ""} \
		{
		set color [option get $win color Color]
		}


	$win.display create rectangle 0 0 0 20 \
		-outline "" -fill $color -tags bar
	$win.display create text [expr {0.5 * $len}] 10 \
		-anchor c -text "0%" -tags value
	return $win
}

proc gauge_value {win val} \
{
	if {$val < 0 || $val > 100} \
		{
		error "bad value \"$val\": should be 0-100"
		}
	set msg [format "%.0f%%" $val]
	$win.display itemconfigure value -text $msg

	set w [expr {0.01 * $val * [winfo width $win.display]}]
	set h [winfo height $win.display]
	$win.display coords bar 0 0 $w $h

	update
}

proc MakeWaitWindow {f cnflag color} \
{
	global CancelXfer

	set CancelXfer 0

	frame $f
	button $f.cancel -text Cancel -command {\
		global CancelXfer; set CancelXfer 1; puts "Canceled"}

	gauge_create $f.g PaleGreen
	option add *Gauge.borderWidth 2 widgetDefault
	option add *Gauge.relief sunken widgetDefault
	option add *Gauge.length 300 widgetDefault
	option add *Gauge.color gray widgetDefault

	pack $f.g -expand yes -fill both \
		-padx 10 -pady 10

	if {$cnflag} \
		{
		pack $f.cancel -side top -padx 3 -pady 3
		}

	

	pack $f
	return $f.g
}

##########################################################
#
# Copy data from radio to template image (a lengthy string).
#
##########################################################
proc Radio2Template { f }\
{
	global FileTypes
	global GlobalParam
	global Home
	global MemFreq
	global MemLabel
	global MemMode
	global ReadRadioFlag

	# OpenDevice

	set msg "Note: Data cannot be transferred "
	append msg "if the radio is in the Preset Mode.\n\n"
	append msg "Instructions (read all steps):\n"
	append msg "1) Ensure the radio is connected to your computer"
	append msg " and powered off.\n"
	append msg "2) Press and hold FUNC and BND keys on"
	append msg " the radio while powering on the radio.\n"
	append msg "3) Click OK.\n"
	append msg "4) After radio is powered on, press MODE.\n"


	set result [tk_dialog .info "Read from radio" \
		$msg \
		info 0 OK Cancel ]

	if {$result} \
		{
		CloseDevice
		return
		}


	# Read memory image from radio.
	if {[ReadImage]} \
		{
		set ReadRadioFlag no

		set msg "Checksum error while reading from radio."
		tk_dialog .error $msg $msg error 0 OK
		CloseDevice
		return
		}
		
	CloseDevice
	set GlobalParam(Populated) 1
	DecodeImage

	set msg "Transfer Complete.\n"
	append msg "Look at the radio display "
	append msg "to view a status message."

	tk_dialog .belch "Read radio" \
		$msg info 0 OK

	set ReadRadioFlag yes

	return
}


##########################################################
#
# Interrogate radio to learn its Preamble
#
# NOTE: This procedure is not finished or used yet.
#
##########################################################
proc InterrogateRadio { f }\
{
	global FileTypes
	global GlobalParam
	global Home
	global MemFreq
	global MemLabel
	global MemMode
	global ReadRadioFlag


	set msg "Note: Data cannot be transferred "
	append msg "if the radio is in the Preset Mode.\n\n"
	append msg "Instructions (read all steps):\n"
	append msg "1) Ensure the radio is connected to your computer"
	append msg " and powered off.\n"
	append msg "2) Press and hold FUNC and BND keys on"
	append msg " the radio while powering on the radio.\n"
	append msg "3) Click OK.\n"
	append msg "4) After radio is powered on, press MODE.\n"


	set result [tk_dialog .info "Interrogate radio" \
		$msg \
		info 0 OK Cancel ]

	if {$result} \
		{
		return
		}


	set line [ReadRx 8]

	binary scan $line "H*" x
	regsub -all ".." $x { \0} x

	set GlobalParam(PreambleRead) $x
	SendAck
	# WritePostamble

	set msg "Interrogation Complete.\n"
	append msg "Look at the radio display "
	append msg "to view a status message."

	tk_dialog .belch "Interrogate radio" \
		$msg info 0 OK


	return
}

##########################################################
# Write memory image to a template file.
##########################################################
proc SaveTemplate { f asflag } \
{
	global GlobalParam
	global TemplateSavedFlag
	global ReadRadioFlag
	global Mimage
	global Nmessages

	if {[string length $Mimage] <= 0} \
		{
		set msg "You must first read template data from"
		append msg " the radio before saving it in a"
		append msg " template file."
		append msg " (Use the Radio menu for reading"
		append msg " from the radio.)"

		tk_dialog .error "No template data" \
			$msg error 0 OK
		return
		}

	set mitypes \
		{
		{"VR-120 template files"           {.tk1}     }
		}

	set filename $GlobalParam(TemplateFilename)

	if { ($GlobalParam(TemplateFilename) == "") \
		|| ($asflag) } \
		{
		set filename \
			[Mytk_getSaveFile $f \
			$GlobalParam(MemoryFileDir) \
			.tk1 \
			"Save VR-120 data to template file" \
			$mitypes]
		}



	if { $filename != "" }\
		{

		if {[ValidateData]} {return}
		MakeWait
		EncodeImage

		# Truncate memory image to the proper length.
		# We want to ignore the several FF records
		# which may have been appended
		# at the end of the image.

		set n [expr {($Nmessages * 32) - 1}]
		set Mimage [string range $Mimage 0 $n]

		KillWait

		set GlobalParam(TemplateFilename) $filename
		SetWinTitle

		set GlobalParam(MemoryFileDir) \
			[ Dirname $GlobalParam(TemplateFilename) ]

		set fid [open $GlobalParam(TemplateFilename) "w"]
		fconfigure $fid -translation binary
		puts -nonewline $fid $Mimage
		close $fid
		set TemplateSavedFlag yes
		}

	return
}


##########################################################
# Read memory image from a template file.
##########################################################
proc OpenTemplate { f } \
{
	global BytesPerMessage
	global GlobalParam
	global Mimage
	global Nmessages

	set mitypes \
		{
		{"VR-120 template files"           {.tk1}     }
		}

	set GlobalParam(TemplateFilename) [Mytk_getOpenFile \
		$f $GlobalParam(MemoryFileDir) \
		"Open template file" $mitypes]


	if { $GlobalParam(TemplateFilename) != "" }\
		{
		set GlobalParam(MemoryFileDir) \
			[ Dirname $GlobalParam(TemplateFilename) ]

		set fid [open $GlobalParam(TemplateFilename) "r"]
		fconfigure $fid -translation binary

		set nbytes [expr {$Nmessages * $BytesPerMessage}]
		set Mimage [read $fid $nbytes]

		close $fid
		set GlobalParam(Populated) 1
		SetWinTitle
		DecodeImage
		}

	return
}


##########################################################
# Import memory channels from a .csv file
##########################################################
proc ImportChannels { f }\
{
	global GlobalParam
	global MemFreq
	global MemLabel
	global MemMode
	global MemPref
	global Mimage
	global RMode


	if {[info exists Mimage] == 0} \
		{
		set msg "You must open a template file\n"
		append msg " or read an image from the radio\n"
		append msg " before importing channels.\n"

		tk_dialog .importinfo "tk120" \
			$msg info 0 OK
		return
		}


	set filetypes \
		{
		{"VR-120 memory channel files"     {.csv .txt}     }
		}


	set filename [Mytk_getOpenFile $f \
		$GlobalParam(MemoryFileDir) \
		"Import channels" $filetypes]

	if {$filename == ""} then {return ""}

	if [ catch { open $filename "r"} fid] \
		{
		# error
		tk_dialog .error "tk120" \
			"Cannot open file $file" \
			error 0 OK

		return
		} 

	# Read entire .csv file at one time.
	set allchannels [read $fid]
	for {set bn 0} {$bn < 10} {incr bn} \
		{
		set ch [expr {$bn * 100}]
		for {set i 0} {$i < 64} {incr i} \
			{
			set MemFreq($ch) 0
			set MemLabel($ch) ""
			set MemMode($ch) NFM
			set MemPref($ch) ""
			incr ch
			}
		}

	set line ""
	set i 0

	# For each line in the .csv file.
	foreach line [split $allchannels "\n" ] \
		{
		update

		incr i
		if { $i > 1 } then\
			{
			# Delete double quote characters.
			regsub -all "\"" $line "" bline
			set line $bline

			if {$line == ""} then {continue}
			
			set msg [ParseCsvLine $line]
			if {$msg != ""} \
				{
				set response [ErrorInFile \
					$msg $line $filename]
				if {$response == 0} then {continue} \
				else {ExitApplication}
				}
			}
		}
		
	close $fid

	return
}

proc ParseCsvLine {line} \
{
	global MemFreq
	global MemLabel
	global MemMode
	global MemPref
	global Mode


	if {$line == ""} {return error}
	set mlist [split $line ","]

	set n [llength $mlist] 
	set m [ expr {4 - $n} ]

	# Add empty fields to the end of the line
	# if there are too few fields.

	for {set i 0} {$i < $m} {incr i} \
		{
		append line ","
		}

	set mlist [split $line ","]

#	if { [llength $mlist] < 4 } \
#		{
#		return "Missing one or more fields."
#		}

	set ch [lindex $mlist 0]
	set freq [lindex $mlist 1]
	set mode [lindex $mlist 2]
	set pref [lindex $mlist 3]
	set label [lindex $mlist 4]


	# VR-120 labels must be upper case.
	set label [string toupper $label]


	if { ($ch < 0) || ($ch > 999) } \
		{
		return "Invalid channel $ch."
		}

	set r [expr { fmod($ch, 100) }]
	set r [expr { int($r) }]

	if {$r > 63} \
		{
		return "Invalid channel $ch."
		}

	
	if { ($freq < 0.0) || ($freq > 2000.0) } \
		{
		return "Invalid frequency $freq."
		}

	if {$mode == ""} {set mode [DefaultMode $freq]}
	set umode [string toupper $mode]

	if { [info exists Mode($umode)] == 0 } \
		{
		return "Invalid mode $mode."
		}

	set MemFreq($ch) $freq
	set MemMode($ch) $mode

	if {$pref != ""} \
		{
		set MemPref($ch) p
		}

	set s [string range $label 0 5]
	set s [format "%-6s" $s]
	set MemLabel($ch) $s


	# puts stderr "$line ch= $ch, freq= $freq, label= $label"
	return ""
}


##########################################################
# Show memory channels in a window.
##########################################################
proc ShowChannels { f }\
{
	global Chvector
	set Chvector ""

	return
}

##########################################################
# Export memory channels to a .csv file
##########################################################
proc ExportChannels { f }\
{
	global FileTypes
	global GlobalParam
	global Home
	global MemFreq
	global MemLabel
	global MemMode
	global MemPref
	global Mimage
	global Ofilename
	global TemplateFilename


	if { ([info exists Mimage] == 0) \
		|| ([string length $Mimage] <= 0) } \
		{
		set msg "You must read data from the radio"
		append msg " before exporting channels."
		append msg " (See the Radio menu.)"
		tk_dialog .info "tk120" $msg info 0 OK
		return
		}


	set Ofilename [Mytk_getSaveFile $f \
		$GlobalParam(MemoryFileDir) \
		.csv \
		"Export memory channels to .csv file" \
		$FileTypes]


	if {$Ofilename != ""} \
		{
		# puts stderr "ExportChannels: Ofilename $Ofilename"
		set GlobalParam(MemoryFileDir) \
			[ Dirname $Ofilename ]

		set fid [open $Ofilename "w"]


		tk_dialog .belch "Export" \
			"Export Complete" info 0 OK

		# Write first line as the field names.
		puts $fid [format "Ch,MHz,Mode,Pref,Label"]

		for {set bn 0} {$bn < 10} {incr bn} \
			{
			set ch [expr {$bn * 100}]
			for {set i 0} {$i < 64} {incr i} \
				{
				if {$MemFreq($ch) <= .000001} \
					{
					incr ch
					continue
					}
	
				set pref ""
				if {$MemPref($ch) == "p"} \
					{
					set pref p
					}
				set s [format "%d,%s,%s,%s," \
					$ch $MemFreq($ch) \
					$MemMode($ch) \
					$pref]
	
				if {$MemLabel($ch) != ""} \
					{
					set l [format "\"%s\"" \
						$MemLabel($ch)]
	
					append s $l
					}
	
				puts $fid $s
				incr ch
				}
			}

		close $fid
		}
	return
}

proc FileExistsDialog { file } \
{
	set result [tk_dialog .fed "Warning" \
		"File $file already exists. Overwrite file?" \
		warning 0 Cancel Overwrite ]

	puts "result is $result"
	return $result
}


##########################################################
# Copy memory image to the radio
##########################################################
proc Image2Radio { f }\
{
	global FileTypes
	global GlobalParam
	global Mimage
	global ReadRadioFlag

	if { $GlobalParam(PreambleRead) == "" } \
		{
		# Do not know Preamble for this radio.
		# This means the user has not read an image
		# from this radio before.

		set msg  ""
		append msg "It appears you have not read an image"
		append msg " from this radio before. "
		append msg " You must first read template data from"
		append msg " your radio before"
		append msg " writing to the radio."

		tk_dialog .error "Write to radio" $msg error 0 OK
		return
		}

	if { ([info exists Mimage] == 0) \
		|| ([string length $Mimage] <= 0)} \
		{
		# No image to write.
		set msg "You must first read template data from"
		append msg " the radio or a file before"
		append msg " writing it to the radio."

		tk_dialog .error "Write to radio" $msg error 0 OK
		return
		}

	# if {$ReadRadioFlag == "yes"}
	if {$ReadRadioFlag == "foo"} \
		{
		# We read an image from the radio.
		# Cannot read from and write to the radio
		# during the same session or else the radio
		# complains. (Reason unknown.)
		#
		# Tell user to save the image file, exit
		# the program, restart the program, read
		# the image file, then write to the radio.

		set msg ""
		append msg "You cannot read from the radio "
		append msg "and write to the radio during the same "
		append msg "session.\n\n"
		append msg "Please:\n"
		append msg "1) Save the memory image in a file,\n"
		append msg "using File --> Save As ...\n"
		append msg "2) Exit this program.\n"
		append msg "3) Restart this program.\n"
		append msg "4) Open the image file you saved "
		append msg "previously, using File --> Open ...\n "
		append msg "5) Then, you can write the image "
		append msg "to the radio."

		tk_dialog .belch "Write blocked warning" \
			$msg warning 0 OK
		return
		}

	if {[ValidateData]} {return}
	MakeWait
	EncodeImage
	KillWait

	# OpenDevice

	set msg "Note: Data cannot be transferred "
	append msg "if the radio is in the Preset Mode.\n\n"
	append msg "Instructions:\n"
	append msg "1) Ensure the radio is connected to"
	append msg " your computer and radio power is off.\n"
	append msg "2) Press and hold FUNC and BND keys on"
	append msg " the radio while"
	append msg " powering on the radio.\n"
	append msg "3) After radio is powered on, press SCAN.\n"
	append msg "4) Click OK or Cancel within a couple of seconds.\n"

	set result [tk_dialog .info "Write to radio" \
		$msg \
		info 0 OK Cancel ]

	if {$result} \
		{
		# User canceled the write.
		CloseDevice
		return
		}


	if { [WriteImage] } \
		{
		set msg "Error while writing to the radio."
		tk_dialog .error "Write error" $msg error 0 OK
		KillWait
		} \
	else \
		{
		set msg "Transfer Complete.\n"
		append msg "Look at the radio display "
		append msg "to view a status message."
		tk_dialog .belch "Transfer Complete" \
			$msg info 0 OK
		}
	CloseDevice

	return
}

###################################################################
# Return 1 if frequency is in range 0 - 2000 exclusive.
###################################################################
proc FreqInRange { f units } \
{
	if {$units == "mhz" } \
		{
		if { $f > 0 && $f < 2000.0 } \
			{
			return 1
			}
		} \
	elseif {$units == "khz" } \
		{
		if { $f > 0 && $f < 2000000.0 } \
			{
			return 1
			}
		}
	return 0
}

###################################################################
# Return 1 if string 's' is a valid frequency.
# Return 0 otherwise.
#
# Units should be khz or mhz
###################################################################
proc CheckFreqValid { s units }\
{
	if {$s == ""} then {return 0}

	# Check for non-digit and non decimal point chars.
	set rc [regexp {^[0-9.]*$} $s]
	if {$rc == 0} then {return 0}


	# All digits.
	set rc [regexp {^[0-9]*$} $s]
	if {$rc == 1} \
		{
		return [FreqInRange $s $units]
		}

	if {$s == "."} then {return 0}

	# Check for Two or more decimal points
	set tmp $s
	set tmp [split $s "."]
	set n [llength $tmp]
	if { $n >= 3 } then {return 0}
	
	return [FreqInRange $s $units]
}


###################################################################
# Set default receiver parameters
###################################################################
proc SetUp { } \
{
	global env
	global GlobalParam
	global RootDir
	global tcl_platform


	if { [regexp "Darwin" $tcl_platform(os) ] } \
		{
		# For Mac OS X.
		set RootDir ":"
		} \
	else \
		{
		set RootDir "/"
		}

	set GlobalParam(Debug) 0
	# set GlobalParam(Device) /dev/ttyS1
	set GlobalParam(Ifilename) {}
	set GlobalParam(MemoryFileDir) $RootDir
	set GlobalParam(PreviousFreq) 0.0

	return
}



###################################################################
# 
# Define receiver parameters before we read the
# global parameter configuration file in case they are missing
# from the configuration file.
# This avoids a tcl error if we tried to refer to an
# undefined variable.
#
# These initial definitions will be overridden with
# definitions from the configuration file.
#
###################################################################

proc PresetGlobals { } \
{
	global GlobalParam
	global Mode
	global Rcfile
	global RootDir
	global tcl_platform

	set GlobalParam(1AOC) 00
	set GlobalParam(1AOE) 00
	set GlobalParam(1A10) 00
	set GlobalParam(1A11) 00
	set GlobalParam(BalloonHelpWindows) on

	set GlobalParam(AntAM) 0
	set GlobalParam(AntFM) 0
	set GlobalParam(Attenuator) 0
	set GlobalParam(BackGroundColor) ""
	set GlobalParam(BankScan) 0
	set GlobalParam(BatterySaver) 1
	set GlobalParam(Beep) 1
	set GlobalParam(Debug) 0
	set GlobalParam(DWorP) 0
	set GlobalParam(ExperimentalTweaks) off
	set GlobalParam(FastTuningStep) 1MHz
	set GlobalParam(FlexStep) 1
	set GlobalParam(FCounter) 50
	set GlobalParam(Flag0Bit0) 0
	set GlobalParam(Flag0Bit1) 0
	set GlobalParam(Flag1Bit0) 0
	set GlobalParam(Flag1Bit6) 0
	set GlobalParam(Flag1Bit7) 0
	set GlobalParam(Font) ""
	set GlobalParam(ForeGroundColor) ""
	set GlobalParam(Lamp) AUTO
	set GlobalParam(LimitSearch) 0
	set GlobalParam(Lock) 0
	set GlobalParam(MemoryFileDir) $RootDir
	set GlobalParam(Mode) $Mode(NFM)
	set GlobalParam(LowestFreq) .1000
	set GlobalParam(HighestFreq) 1300.0000
	set GlobalParam(PreambleRead) ""
	set GlobalParam(Resume) 5
	set GlobalParam(SetMenuItem) 0
	set GlobalParam(SMeterBeep) 0
	set GlobalParam(SMeterSens) 0
	set GlobalParam(SetMenuItem) 0
	set GlobalParam(SortBank) -1
	set GlobalParam(SortType) "label"
	set GlobalParam(SSContinue) 0
	set GlobalParam(FlagByte3) 0
	set GlobalParam(TroughColor) ""
	set GlobalParam(TuningStep) 5
	set GlobalParam(VFOFreq) 162.4000
	set GlobalParam(Wadjust) 9
	set GlobalParam(WhichDW) 0
	set GlobalParam(WhichModel) USA
	set GlobalParam(WXFreq) 162.55
	set GlobalParam(WXMode) AUTO

	return
}


###################################################################
# Set global variables after reading the global
# configuration file so these settings override
# whatever values were in the configuration file.
###################################################################

proc OverrideGlobals { } \
{
	global env
	global GlobalParam
	global RootDir
	global tcl_platform


	set GlobalParam(BypassAllEncoding) 0
	set GlobalParam(Ifilename) {}
	set GlobalParam(NmsgsRead) 0
	set GlobalParam(Populated) 0
	set GlobalParam(PostambleRead) ""
	set GlobalParam(TemplateFilename) {}
	set GlobalParam(UserPort) 0
	set GlobalParam(Wadjust) 9

	# Note on MacOS X:
	# The initial directory passed to the file chooser widget.
	# The problem here is that osx's tcl is utterly busted.
	# The _only_ pathname it accepts is ':' - no other ones work.
	# Now this isn't as bad as you might think because
	# the native macos file selector widget persistantly
	# remembers the last place you opened/saved a file
	# for a particular application. So the logic to
	# remember this is simply redundant on macos anyway...
	# Presumably they'll fix this someday and we can take
	# out the hack.
	# - Ben Mesander

	if { [regexp "Darwin" $tcl_platform(os) ] } \
		{
		# kluge for MacOS X.

		set GlobalParam(LogFileDir) $RootDir
		set GlobalParam(MemoryFileDir) $RootDir

		if {$GlobalParam(Ifilename) != ""} \
			{
			set GlobalParam(Ifilename) $RootDir
			}
		}

	return
}

###################################################################
#
# Parse data from inside the memory image and store it
# in global arrays.
#
###################################################################
proc DecodeImage { } \
{
	MakeWait
	update idletasks

	DecodeMisc
	DecodeTweaks
	DecodeScanResume
	DecodeSmartSearch
	DecodeMemories
	DecodeSearchBanks
	DecodeBandEdges
	DecodeBandStack
	DecodeSkipFreqs
	DecodeDWBanks
	DecodePresets
	DecodeAutoBand
	DecodeBankLinks

	update idletasks
	KillWait
	update idletasks
	return
}

###################################################################
#
# Parse data from inside the memory image and store it
# in global arrays.
#
###################################################################
proc DecodeMisc { } \
{
	global GlobalParam
	global ImageAddr
	global Mimage
	global Priority
	global PriorityMode
	global RBatterySaver
	global RCounter
	global RMode
	global RLamp
	global RFstep
	global RStep



	# Parse VFO frequency
	scan $ImageAddr(VFOFreq) "%x" first
	set last [expr {$first + 3}]
	set s [string range $Mimage $first $last]
	set f [BCD2Freq3p $s]
	set GlobalParam(VFOFreq) $f


	# Parse VFO mode.
	scan $ImageAddr(VFOMode) "%x" first

	set last $first
	set s [string range $Mimage $first $last]

	binary scan $s "H2" s
	set m [string index $s 1]
	if { [info exists RMode($m)] } \
		{
		set GlobalParam(VFOMode) [string toupper $RMode($m)]
		} \
	else \
		{
		set GlobalParam(VFOMode)
		}


	# Parse priority frequency
	scan $ImageAddr(Priority) "%x" first

	set last [expr {$first + 3}]
	set s [string range $Mimage $first $last]
	set Priority [BCD2Freq3p $s]

	# Parse priority mode.
	scan $ImageAddr(PriorityMode) "%x" first

	set last $first
	set s [string range $Mimage $first $last]

	binary scan $s "H2" s
	set m [string index $s 1]
	if { [info exists RMode($m)] } \
		{
		set PriorityMode [string toupper $RMode($m)]
		} \
	else \
		{
		set PriorityMode {?}
		}


	# Parse Scan Resume time
	scan $ImageAddr(Resume) "%x" first
	set s [string range $Mimage $first $first]
	if {$s == ""} \
		{
		set GlobalParam(Resume) 1
		} \
	else \
		{
		set c [Char2Int $s]
		set GlobalParam(Resume) [expr {$c + 1}]
		}

	# Parse Attenuator, on or off flag
	scan $ImageAddr(FlagByte3) "%x" first
	set byte [string range $Mimage $first $first]
	set GlobalParam(Attenuator) [GetBit $byte 5]

	# Parse AM Antenna bit
	scan $ImageAddr(FlagByte0) "%x" first
	set s [string range $Mimage $first $first]
	set GlobalParam(AntAM) [GetBit $s 7]

	# Parse FM Antenna bit
	scan $ImageAddr(FlagByte1) "%x" first
	set s [string range $Mimage $first $first]
	set GlobalParam(AntFM) [GetBit $s 5]

	# Parse Keypad Lock, on or off flag
	scan $ImageAddr(FlagByte3) "%x" first
	set byte [string range $Mimage $first $first]
	set GlobalParam(Lock) [GetBit $byte 1]

	# Parse battery saver interval
	scan $ImageAddr(BatterySaver) "%x" first
	set byte [string range $Mimage $first $first]
	set c [Char2Int $byte]
	if { [info exists RBatterySaver($c)] } \
		{
		set GlobalParam(BatterySaver) $RBatterySaver($c)
		} \
	else \
		{
		set GlobalParam(BatterySaver) 1
		}


	# Parse bank scan flag
	scan $ImageAddr(FlagByte0) "%x" first
	set byte [string range $Mimage $first $first]
	set GlobalParam(BankScan) [GetBit $byte 4]


	# Parse beep tone, on or off flag
	scan $ImageAddr(FlagByte0) "%x" first
	set byte [string range $Mimage $first $first]
	set GlobalParam(Beep) [GetBit $byte 5]

	# parse S-Meter beep alarm, on or off flag
	scan $ImageAddr(FlagByte1) "%x" first
	set byte [string range $Mimage $first $first]
	set GlobalParam(SMeterBeep) [GetBit $byte 2]

	# parse Decrease S-Meter sensitivity, on or off flag
	scan $ImageAddr(FlagByte1) "%x" first
	set byte [string range $Mimage $first $first]
	set GlobalParam(SMeterSens) [GetBit $byte 1]

	# parse Tuning Step
	scan $ImageAddr(VFOStep) "%x" first
	set byte [string range $Mimage $first $first]
	binary scan $byte "H2" s

	if { [info exists RStep($s)] } \
		{
		set GlobalParam(TuningStep) $RStep($s)
		} \
	else \
		{
		set GlobalParam(TuningStep) AUTO
		}

	# Parse Freq Counter width value
	scan $ImageAddr(FCounter) "%x" first
	set s [string range $Mimage $first $first]
	set n [Char2Int $s]
	if { [info exists RCounter($n)] } \
		{
		set GlobalParam(FCounter) $RCounter($n)
		} \
	else \
		{
		set GlobalParam(FCounter) $RCounter(0)
		}

	# Parse VFO/Limit Search bit
	scan $ImageAddr(FlagByte0) "%x" first
	set s [string range $Mimage $first $first]
	set GlobalParam(LimitSearch) [GetBit $s 3]


	# Parse 2 Lamp bits
	scan $ImageAddr(Lamp) "%x" first
	set s [string range $Mimage $first $first]
	set bit6 [GetBit $s 6]
	set bit7 [GetBit $s 7]
	set n [expr {(2 * $bit6) + $bit7}]
	set GlobalParam(Lamp) $RLamp($n)

	# Parse Fast Tuning Step
	scan $ImageAddr(FastTuningStep) "%x" first
	set s [string range $Mimage $first $first]
	set n [Char2Int $s]
	if { [info exists RFstep($n)] } \
		{
		set GlobalParam(FastTuningStep) $RFstep($n)
		} \
	else \
		{
		set GlobalParam(FastTuningStep) 1MHz
		}

	# Parse DW or Priority flag
	scan $ImageAddr(FlagByte0) "%x" first
	set s [string range $Mimage $first $first]
	set GlobalParam(DWorP) [GetBit $s 2]


	# Parse Flex Step bit
	scan $ImageAddr(FlagByte0) "%x" first
	set s [string range $Mimage $first $first]
	set GlobalParam(FlexStep) [GetBit $s 6]


	# Parse Which DW bank
	scan $ImageAddr(WhichDW) "%x" first
	set s [string range $Mimage $first $first]
	if {$s == ""} \
		{
		set GlobalParam(WhichDW) 0
		} \
	else \
		{
		set c [Char2Int $s]
		set GlobalParam(WhichDW) $c
		}


	return
}


###################################################################
#
# Parse data from inside the memory image and store it
# in global arrays.
#
###################################################################
proc DecodeTweaks { } \
{
	global GlobalParam
	global ImageAddr
	global Mimage


	# Parse Set Menu Item
	scan $ImageAddr(SetMenuItem) "%x" first
	set s [string range $Mimage $first $first]
	set GlobalParam(SetMenuItem) [Char2Int $s]

	# Parse User Port tweak, on or off flag
	scan $ImageAddr(FlagByte1) "%x" first
	set byte [string range $Mimage $first $first]
	set GlobalParam(UserPort) [GetBit $byte 3]

	# Parse experimental bits, on or off flag. Affect unknown.
	scan $ImageAddr(FlagByte0) "%x" first
	set byte [string range $Mimage $first $first]
	set GlobalParam(Flag0Bit0) [GetBit $byte 0]

	# Parse experimental bits, on or off flag. Affect unknown.
	scan $ImageAddr(FlagByte0) "%x" first
	set byte [string range $Mimage $first $first]
	set GlobalParam(Flag0Bit1) [GetBit $byte 1]

	# Parse experimental bits, on or off flag. Affect unknown.
	scan $ImageAddr(FlagByte1) "%x" first
	set byte [string range $Mimage $first $first]
	set GlobalParam(Flag1Bit0) [GetBit $byte 0]

	# Parse experimental bits, on or off flag. Affect unknown.
	scan $ImageAddr(FlagByte1) "%x" first
	set byte [string range $Mimage $first $first]
	set GlobalParam(Flag1Bit6) [GetBit $byte 6]

	# Parse experimental bits, on or off flag. Affect unknown.
	scan $ImageAddr(FlagByte1) "%x" first
	set byte [string range $Mimage $first $first]
	set GlobalParam(Flag1Bit7) [GetBit $byte 7]

	# Parse 1A11
	scan $ImageAddr(FlagByte3) "%x" first
	set s [string range $Mimage $first $first]
	set t [Char2Int $s]
	set GlobalParam(1A11) [format "%02x" $t]

	scan $ImageAddr(1A0C) "%x" first
	set s [string range $Mimage $first $first]
	set t [Char2Int $s]
	set GlobalParam(1A0C) [format "%02x" $t]

	scan $ImageAddr(1A0E) "%x" first
	set s [string range $Mimage $first $first]
	set t [Char2Int $s]
	set GlobalParam(1A0E) [format "%02x" $t]

	scan $ImageAddr(1A10) "%x" first
	set s [string range $Mimage $first $first]
	set t [Char2Int $s]
	set GlobalParam(1A10) [format "%02x" $t]


	return
}

###################################################################
#
# Parse data from inside the memory image and store it
# in global arrays.
#
# Notes:
#	Bank links are bit flags.
#	If a bank is marked to be scanned, its corresponding bit
#	is 1.
#
#	There are 10 bank links, one for each memory bank.
#	The 10 flags are stored in 2 consecutive bytes.
#	Flags for banks 0 - 7 are stored in the first byte.
#	Bank 0 flag is the least significant bit.
#	Bank 7 flag is the most significant bit.
#
#	Flags for banks 8 - 9 are stored in the second byte.
#	Bank 9 flag is the least significant bit.
#
###################################################################
proc DecodeBankLinks { } \
{
	global BankLink
	global ImageAddr
	global Mimage


	# Parse linked banks
	scan $ImageAddr(BankLinks) "%x" first

	set byte [string range $Mimage $first $first]

	for {set bn 0} {$bn < 8} {incr bn} \
		{
		set i [expr {7 - $bn}]
		set BankLink($bn) [GetBit $byte $i]
		}

	incr first
	set byte [string range $Mimage $first $first]

	for {set bn 8} {$bn < 10} {incr bn} \
		{
		set i [expr {7 - ($bn - 8)}]
		set BankLink($bn) [GetBit $byte $i]
		}

	return
}


###################################################################
#
# Parse data from inside the memory image and store it
# in global arrays.
#
# NOTES:
#	Each memory channel is represented by 4 consecutive bytes.
#	The first 3 bytes contain the frequency digits in hex.
#	The most significant nibble in the first byte is 0-E.
#	Frequencies of 1000 MHz and higher start with a letter.
#
#	The fourth byte is a little strange. It contains both the mode
#	(AM, NFM, or WFM) and at least one or more bit flags.
#	The 2 least significant bits contain the mode.
#
#	The bit flags indicate whether to the frequency is 500 Hz
#	higher (bit 1) or 250 Hz higher (bit 2) than the number
#	indicated by the preceding 3 bytes.
#	
###################################################################
proc DecodeMemories { } \
{
	global Mimage
	global MemFreq
	global MemLabel
	global MemMode
	global MemPref
	global ImageAddr
	global RMode

	# Parse memory channel text labels.
	scan $ImageAddr(MemoryLabels) "%x" first
	set last [expr {$first + 5}]

	for {set bn 0} {$bn < 10} {incr bn} \
		{
		set ch [expr {$bn * 100}]
		for {set i 0} {$i < 64} {incr i} \
			{
			set s [string range $Mimage $first $last]
			set s [string trimright $s " "]
			set MemLabel($ch) $s
			incr first 6
			incr last 6
			incr ch
			}
		}


	# Parse memory channel frequencies.
	scan $ImageAddr(MemoryFreqs) "%x" first

	for {set bn 0} {$bn < 10} {incr bn} \
		{
		set ch [expr {$bn * 100}]
		for {set i 0} {$i < 64} {incr i} \
			{
			set last [expr {$first + 3}]
			set s [string range $Mimage $first $last]

			set f [BCD2Freq3p $s]

			set MemFreq($ch) $f
			incr first 4
			incr last 4
			incr ch
			}
		}


	# Parse memory channel modes.
	scan $ImageAddr(MemoryModes) "%x" first

	for {set bn 0} {$bn < 10} {incr bn} \
		{
		set ch [expr {$bn * 100}]
		for {set i 0} {$i < 64} {incr i} \
			{
			set byte [string range $Mimage $first $first]
	
			binary scan $byte "H2" s
			# set m [string index $s 1]

			# Two LS bits denote mode.
			set bit6 [GetBit $byte 6]
			set bit7 [GetBit $byte 7]

			set m [expr {$bit6 + $bit6 + $bit7}]

			if { [info exists RMode($m)] } \
				{
				set MemMode($ch) $RMode($m)
				} \
			else \
				{
				set MemMode($ch) {?}
				}
	
			if {$MemFreq($ch) >= 0.100} \
				{
				# Preferential flag is bit 5.
				# If the bit is set 
				# this means the channel
				# is a Preferential channel.
		
				set pref [GetBit $byte 5]
# puts stderr "ch $ch, $MemFreq($ch), $s, first $first, pref $pref, $MemMode($ch)"
		
				set MemPref($ch) ""
				if {$pref} \
					{
					set MemPref($ch) "p"
					}
		
				} \
			else \
				{
				# Empty channel
				set MemPref($ch) ""
				}
			incr first 4
			incr ch
			}
		}

	return
}


###################################################################
#
# Parse data from inside the memory image and store it
# in global arrays.
#
###################################################################
proc DecodeSearchBanks { } \
{
	global LimitScan
	global Mimage
	global MemFreq
	global MemLabel
	global MemMode
	global MemPref
	global ImageAddr
	global RMode


	# Parse lower limit frequencies.
	scan $ImageAddr(SearchFreqFirst) "%x" first

	for {set bn 0} {$bn < 9} {incr bn} \
		{
		if {$bn == 8} \
			{
			scan $ImageAddr(SearchFreqFirstD) "%x" first
			}
		set last [expr {$first + 3}]
		set s [string range $Mimage $first $last]

		set f [BCD2Freq3p $s]

		set LimitScan($bn,lower) $f
		incr first 4
		}


	# Parse upper limit frequencies.
	scan $ImageAddr(SearchFreqSecond) "%x" first

	for {set bn 0} {$bn < 9} {incr bn} \
		{
		if {$bn == 8} \
			{
			scan $ImageAddr(SearchFreqSecondD) "%x" first
			}
		set last [expr {$first + 3}]
		set s [string range $Mimage $first $last]

		set f [BCD2Freq3p $s]

		set LimitScan($bn,upper) $f
		incr first 4
		}


	# Parse lower edge modes.
	scan $ImageAddr(SearchModeFirst) "%x" first

	for {set bn 0} {$bn < 9} {incr bn} \
		{
		if {$bn == 8} \
			{
			scan $ImageAddr(SearchModeFirstD) "%x" first
			}
		set last $first
		set s [string range $Mimage $first $last]

		binary scan $s "H2" s
		set m [string index $s 1]
		if { [info exists RMode($m)] } \
			{
			set LimitScan($bn,lmode) \
				[string toupper $RMode($m)]
			} \
		else \
			{
			set LimitScan($bn,lmode) {?}
			}

		set LimitScan($bn,link) 0

		incr first 4
		}


	# Parse upper edge modes.
	scan $ImageAddr(SearchModeSecond) "%x" first

	for {set bn 0} {$bn < 9} {incr bn} \
		{
		if {$bn == 8} \
			{
			scan $ImageAddr(SearchModeSecondD) "%x" first
			}
		set last $first
		set s [string range $Mimage $first $last]

		binary scan $s "H2" s
		set m [string index $s 1]
		if { [info exists RMode($m)] } \
			{
			set LimitScan($bn,umode) \
				[string toupper $RMode($m)]
			} \
		else \
			{
			set LimitScan($bn,umode) {?}
			}

		set LimitScan($bn,link) 0

		incr first 4
		}

	return
}


###################################################################
#
# Parse data from inside the memory image and store it
# in global arrays.
#
# NOTES:
#	Each Skip channel is composed of a 3-byte frequency
#	followed by a special 4th byte.
#	The 4th byte contains the splinter flags in bits
#	1 and 2, and the mode in the right nibble.
###################################################################
proc DecodeSkipFreqs { } \
{
	global Mimage
	global ImageAddr
	global RMode
	global Skip


	# Parse VFO skip frequencies.
	scan $ImageAddr(SkipFreqs) "%x" first
	set last [expr {$first + 3}]

	for {set bn 0} {$bn < 64} {incr bn} \
		{
		set s [string range $Mimage $first $last]

		set f [BCD2Freq3p $s]

		set Skip($bn,freq) $f
		incr first 4
		incr last 4
		}

	# Parse VFO skip modes.
	# The mode is contained within the right nibble
	# of a byte.

	scan $ImageAddr(SkipModes) "%x" first

	for {set bn 0} {$bn < 64} {incr bn} \
		{
		set s [string range $Mimage $first $first]

		binary scan $s "H2" m
		set m [string index $m 1]

		if { [info exists RMode($m)] } \
			{
			set mode $RMode($m)

			set Skip($bn,mode) \
				[string toupper $mode]
			} \
		else \
			{
			set Skip($bn,mode) NFM
			}
		incr first 4
		}

	return
}



###################################################################
#
# Parse data from inside the memory image and store it
# in global arrays.
#
# NOTES:
#	Each band edge channel is composed of a 3-byte frequency
#	followed by a special 4th byte.
#	The 4th byte contains the splinter flags in bits
#	1 and 2, and the mode in the right nibble.
###################################################################
proc DecodeBandEdges { } \
{
	global Mimage
	global ImageAddr
	global RMode
	global BandEdge


	# Parse Band Edge frequencies.
	scan $ImageAddr(BandEdgeFreq) "%x" first
	set last [expr {$first + 3}]

	for {set bn 0} {$bn < 12} {incr bn} \
		{
		set s [string range $Mimage $first $last]

		set f [BCD2Freq3p $s]

		set BandEdge($bn,freq) $f
		incr first 4
		incr last 4
		}

	# Parse Band Edge modes.
	# The mode is contained within the right nibble
	# of a byte.

	scan $ImageAddr(BandEdgeMode) "%x" first

	for {set bn 0} {$bn < 12} {incr bn} \
		{
		set s [string range $Mimage $first $first]

		binary scan $s "H2" m
		set m [string index $m 1]

		if { [info exists RMode($m)] } \
			{
			set mode $RMode($m)

			set BandEdge($bn,mode) \
				[string toupper $mode]
			} \
		else \
			{
			set BandEdge($bn,mode) NFM
			}
		incr first 4
		}

	return
}


###################################################################
#
# Parse data from inside the memory image and store it
# in global arrays.
#
###################################################################
proc DecodeDWBanks { } \
{
	global DualWatch
	global Mimage
	global ImageAddr
	global RMode


	# Parse lower limit frequencies.
	scan $ImageAddr(DualWatchFirst) "%x" first

	for {set bn 0} {$bn < 9} {incr bn} \
		{
		if {$bn == 8} \
			{
			scan $ImageAddr(DualWatchFirstD) "%x" first
			}
		set last [expr {$first + 3}]
		set s [string range $Mimage $first $last]

		set f [BCD2Freq3p $s]

		set DualWatch($bn,first) $f
		incr first 4
		}


	# Parse upper limit frequencies.
	scan $ImageAddr(DualWatchSecond) "%x" first

	for {set bn 0} {$bn < 9} {incr bn} \
		{
		if {$bn == 8} \
			{
			scan $ImageAddr(DualWatchSecondD) "%x" first
			}
		set last [expr {$first + 3}]
		set s [string range $Mimage $first $last]

		set f [BCD2Freq3p $s]

		set DualWatch($bn,second) $f
		incr first 4
		}


	# Parse modes.
	scan $ImageAddr(DualWatchFirstMode) "%x" first

	for {set bn 0} {$bn < 9} {incr bn} \
		{
		if {$bn == 8} \
			{
			scan $ImageAddr(DualWatchFirstModeD) "%x" first
			}
		set last $first
		set s [string range $Mimage $first $last]

		binary scan $s "H2" s
		set m [string index $s 1]
		if { [info exists RMode($m)] } \
			{
			set DualWatch($bn,mode1) \
				[string toupper $RMode($m)]
			} \
		else \
			{
			set DualWatch($bn,mode1) {?}
			}
		incr first 4
		}



	# Parse modes.
	scan $ImageAddr(DualWatchSecondMode) "%x" first

	for {set bn 0} {$bn < 9} {incr bn} \
		{
		if {$bn == 8} \
			{
			scan $ImageAddr(DualWatchSecondModeD) "%x" first
			}
		set last $first
		set s [string range $Mimage $first $last]

		binary scan $s "H2" s
		set m [string index $s 1]
		if { [info exists RMode($m)] } \
			{
			set DualWatch($bn,mode2) \
				[string toupper $RMode($m)]
			} \
		else \
			{
			set DualWatch($bn,mode2) {?}
			}

		incr first 4
		}

	return
}



###################################################################
#
# Parse data from inside the memory image and store it
# in global arrays.
#
###################################################################
proc DecodePresets { } \
{
	global GlobalParam
	global ImageAddr
	global Mimage
	global Preset
	global RMode


	# Parse preset frequencies.
	scan $ImageAddr(PresetFreqs) "%x" first

	for {set bn 0} {$bn < 11} {incr bn} \
		{
		set last [expr {$first + 3}]
		set s [string range $Mimage $first $last]

		set f [BCD2Freq3p $s]

		set Preset($bn,freq) $f

		incr first 4
		}


	# Parse preset modes.
	scan $ImageAddr(PresetModes) "%x" first

	for {set bn 0} {$bn < 11} {incr bn} \
		{
		set m [string range $Mimage $first $first]

		binary scan $m "H2" mode
		set nmode [string index $mode 1]

		if { [info exists RMode($nmode)] } \
			{
			set Preset($bn,mode) $RMode($nmode)
			} \
		else \
			{
			set Preset($bn,mode) "?"
			}
		incr first 4
		}


	# Parse WX frequency
	scan $ImageAddr(WXFreq) "%x" first
	set last [expr {$first + 3}]
	set s [string range $Mimage $first $last]
	set f [BCD2Freq3p $s]
	set GlobalParam(WXFreq) $f


	# Parse WX mode.
	scan $ImageAddr(WXMode) "%x" first

	set last $first
	set s [string range $Mimage $first $last]

	binary scan $s "H2" s
	set m [string index $s 1]
	if { [info exists RMode($m)] } \
		{
		set GlobalParam(WXMode) [string toupper $RMode($m)]
		} \
	else \
		{
		set GlobalParam(WXMode)
		}


	return
}

###################################################################
#
# Parse data from inside the memory image and store it
# in global arrays.
#
###################################################################
proc DecodeBandStack { } \
{
	global GlobalParam
	global ImageAddr
	global Mimage
	global BandStack
	global RMode


	# Parse BandStack frequencies.
	scan $ImageAddr(BandStackFreq) "%x" first

	for {set bn 0} {$bn < 12} {incr bn} \
		{
		set last [expr {$first + 3}]
		set s [string range $Mimage $first $last]

		set f [BCD2Freq3p $s]

		set BandStack($bn,freq) $f

		incr first 4
		}


	# Parse BandStack modes.
	scan $ImageAddr(BandStackMode) "%x" first

	for {set bn 0} {$bn < 12} {incr bn} \
		{
		set m [string range $Mimage $first $first]

		binary scan $m "H2" mode
		set nmode [string index $mode 1]

		if { [info exists RMode($nmode)] } \
			{
			set BandStack($bn,mode) $RMode($nmode)
			} \
		else \
			{
			set BandStack($bn,mode) "?"
			}
		incr first 4
		}


	return
}


###################################################################
#
# Parse data from inside the memory image and store it
# in global arrays.
#
###################################################################
proc DecodeAutoBand { } \
{
	global AutoBand
	global AutoBandMode
	global Mimage
	global ImageAddr
	global RMode


	# Parse autoband frequencies.
	scan $ImageAddr(AutoBand) "%x" first

	for {set bn 1} {$bn < 40} {incr bn} \
		{
		set last [expr {$first + 3}]
		set s [string range $Mimage $first $last]

		set f [BCD2Freq3a $s]

		set AutoBand($bn,freq) $f
		incr first 4
		}


	# Parse modes.
	scan $ImageAddr(AutoBandSM) "%x" first

	for {set bn 1} {$bn < 40} {incr bn} \
		{
		set last $first
		set s [string range $Mimage $first $last]

		set i $bn
		incr i -1

		binary scan $s "H2" m
		if { [info exists AutoBandMode($m)] } \
			{
			set lst [split $AutoBandMode($m) {,}]
			set mode [lindex $lst 0]
			set step [lindex $lst 1]

#			puts stderr "$bn  $AutoBand($bn,freq) $m $mode $step"
			set AutoBand($i,mode) \
				[string toupper $mode]

			set AutoBand($i,step) $step
			} \
		else \
			{
			set AutoBand($i,mode) {?}
			set AutoBand($i,step) {?}
			}
		incr first 4
		}

#	for {set i 1} {$i < 40} {incr i} \
#		{
#		puts stderr "$i) $AutoBand($i,freq), $AutoBand($i,step), $AutoBand($i,mode)"
#		}

	return
}


###################################################################
#
# Parse data from inside the memory image and store it
# in global arrays.
#
###################################################################
proc DecodeSmartSearch { } \
{
	global GlobalParam
	global Mimage
	global ImageAddr
	global SmartSearch


	# Parse Smart Search frequencies.

	scan $ImageAddr(SmartSearchL) "%x" first
	set last [expr {$first + 3}]
	set s [string range $Mimage $first $last]
	set f [BCD2Freq3p $s]
	set SmartSearch(lower) $f

	scan $ImageAddr(SmartSearchU) "%x" first
	set last [expr {$first + 3}]
	set s [string range $Mimage $first $last]
	set f [BCD2Freq3p $s]
	set SmartSearch(upper) $f

	scan $ImageAddr(SmartSearchS) "%x" first
	set last [expr {$first + 3}]
	set s [string range $Mimage $first $last]
	set f [BCD2Freq3p $s]
	set SmartSearch(start) $f

	# Parse Single/Continue bit
	scan $ImageAddr(FlagByte1) "%x" first
	set s [string range $Mimage $first $first]
	set GlobalParam(SSContinue) [GetBit $s 4]

	return
}


###################################################################
# Decode a 4 byte BCD frequency.
#
# Returns:	frequency in MHz  * 100000
###################################################################

proc BCD2Freq { s } \
{
	global GlobalParam
	

	# Note: Yaesu packs two digits per byte, one per nibble.
	# An important exception is the most significant nibble
	# in the most significant byte.  That nibble can be
	# 0-9 or a-f.
	# a-f means 10-15.

	# Frequency digit pairs.

	set i 0

	set f1 [string index $s $i]
	# binary scan $f1 "H2" f1
	set f1 [Char2Int $f1]
	set f1 [format "%02x" $f1]
	regsub -nocase {a} $f1 10 f1
	regsub -nocase {b} $f1 11 f1
	regsub -nocase {c} $f1 12 f1
	regsub -nocase {d} $f1 13 f1
	regsub -nocase {e} $f1 14 f1
	regsub -nocase {f} $f1 15 f1
	incr i

	set f2 [string index $s $i]
	binary scan $f2 "H2" f2
	incr i

	set f3 [string index $s $i]
	binary scan $f3 "H2" f3
	incr i

	set f4 [string index $s $i]
	binary scan $f4 "H2" f4
	incr i

	set f [format "%s%s%s%s" $f1 $f2 $f3 $f4]
	set f [string trimleft $f 0]

	if { $f == ""} \
		{
		set f "00000000"
		}

	# Check for non-digit chars.
	set rc [regexp {^[0-9]*$} $f]
	if {$rc == 0} then {set f "00000000"}

	set f [expr {($f/100000.0) + .000005}]
	set f [ format "%.4f" $f]
	return $f
}


###################################################################
# Decode a 3 byte BCD frequency.
#
# Returns:	frequency in MHz  * 100000
###################################################################

proc BCD2Freq3 { s } \
{
	global GlobalParam
	

	# Note: Yaesu packs two digits per byte, one per nibble.
	# An important exception is the most significant nibble
	# in the most significant byte.  That nibble can be
	# 0-9 or a-f.
	# a-f means 10-15.

	# Frequency digit pairs.

	set i 0

	set f1 [string index $s $i]
	binary scan $f1 "H2" f1
	regsub -nocase {a} $f1 10 f1
	regsub -nocase {b} $f1 11 f1
	regsub -nocase {c} $f1 12 f1
	regsub -nocase {d} $f1 13 f1
	regsub -nocase {e} $f1 14 f1
	regsub -nocase {f} $f1 15 f1
	incr i

	set f2 [string index $s $i]
	binary scan $f2 "H2" f2
	incr i

	set f3 [string index $s $i]
	binary scan $f3 "H2" f3
	incr i


	set f [format "%s%s%s" $f1 $f2 $f3]
	set f [string trimleft $f 0]

	if { $f == ""} \
		{
		set f "00000000"
		}

	# Check for non-digit chars.
	set rc [regexp {^[0-9]*$} $f]
	if {$rc == 0} then {set f "00000000"}

	set f [expr {$f/1000.0}]
	set f [ format "%.4f" $f]
	return $f
}


###################################################################
# Decode a 3+ byte BCD frequency.
#
# Returns:	frequency in MHz  * 100000
###################################################################

proc BCD2Freq3p { s } \
{
	global GlobalParam
	

	# Note: Yaesu packs two digits per byte, one per nibble.
	# An important exception is the most significant nibble
	# in the most significant byte.  That nibble can be
	# 0-9 or a-f.
	# a-f means 10-15.
	# Bit 1 in the last byte signifies a splinter channel
	# and means add 0.0005 MHz (500 Hz).

	# Frequency digit pairs.

	set i 0

	set f1 [string index $s $i]
	binary scan $f1 "H2" f1
	regsub -nocase {a} $f1 10 f1
	regsub -nocase {b} $f1 11 f1
	regsub -nocase {c} $f1 12 f1
	regsub -nocase {d} $f1 13 f1
	regsub -nocase {e} $f1 14 f1
	regsub -nocase {f} $f1 15 f1
	incr i

	set f2 [string index $s $i]
	binary scan $f2 "H2" f2
	incr i

	set f3 [string index $s $i]
	binary scan $f3 "H2" f3
	incr i

	set f [format "%s%s%s" $f1 $f2 $f3]
	set f [string trimleft $f 0]

	if { $f == ""} \
		{
		# puts stderr "BCD2Freq3p: 0 frequency"
		set f 0
		}


	# Check for non-digit chars.
	set rc [regexp {^[0-9]*$} $f]
	if {$rc == 0} \
		{
		set f 0
		}

	set f [expr {$f/1000.0}]

	set f4 [string index $s $i]

	if { ($f > .001) && ([GetBit $f4 1])} \
		{
		set f [expr {$f + .0005}]
		} \
	elseif { ($f > .001) && ([GetBit $f4 2])} \
		{
		set f [expr {$f + .00025}]
		}

	set f [ format "%.5f" $f]
	return $f
}



###################################################################
# Decode a 3 byte BCD frequency for Auto Band
#
# Returns:	frequency in MHz
###################################################################

proc BCD2Freq3a { s } \
{
	global GlobalParam
	

	# Note: Yaesu packs two digits per byte, one per nibble.
	# An important exception is the most significant nibble
	# in the most significant byte.  That nibble can be
	# 0-9 or a-f.
	# a-f means 10-15.

	# Frequency digit pairs.

	set i 0

	set f1 [string index $s $i]
	binary scan $f1 "H2" f1
	regsub -nocase {a} $f1 10 f1
	regsub -nocase {b} $f1 11 f1
	regsub -nocase {c} $f1 12 f1
	regsub -nocase {d} $f1 13 f1
	regsub -nocase {e} $f1 14 f1
	regsub -nocase {f} $f1 15 f1
	incr i

	set f2 [string index $s $i]
	binary scan $f2 "H2" f2
	incr i

	set f3 [string index $s $i]
	binary scan $f3 "H2" f3
	incr i

	set f [format "%s%s%s" $f1 $f2 $f3]
	set f [string trimleft $f 0]

	if { $f == ""} \
		{
		set f "00000000"
		}

	# Check for non-digit chars.
	set rc [regexp {^[0-9]*$} $f]
	if {$rc == 0} then {set f "00000000"}

	set f [ expr { $f / 1000.0} ]
	set f [ format "%.4f" $f]
	return $f
}



###################################################################
# Encode a frequency into Yaesu 4 byte format.
#
# Input:	frequency in MHz
# Returns:	4 byte BCD Yaesu coded frequency
###################################################################
proc Freq2BCD { f } \
{
	global GlobalParam
	
	# Note: Yaesu packs two digits per byte, one per nibble.
	# An important exception is the most significant nibble
	# in the most significant byte.  That nibble can be
	# 0-9 or a-f.
	# a-f means 10-15.


	# Check for non-digit chars.
	set rc [regexp {^[0-9\.]*$} $f]
	if {$rc == 0} {set f 0.0}

	if { ($f > $GlobalParam(HighestFreq)) \
		|| ($f < $GlobalParam(LowestFreq)) } \
		{
		set f 0.0
		}

	set z [expr {($f * 100000.0) + .000005}]
	set z [expr {int($z)}]

	if {$f <= 1000.0} \
		{
		set z [format "%08d" $z]

		set s [string range $z 0 1]
		set b1 [binary format "H2" $s]

		set i 2
		set j 3
		} \
	else \
		{
		set z [format "%09d" $z]
		set s [string range $z 0 1]


		regsub -nocase {10} $s "a" s
		regsub -nocase {11} $s "b" s
		regsub -nocase {12} $s "c" s
		regsub -nocase {13} $s "d" s
		regsub -nocase {14} $s "e" s
		regsub -nocase {15} $s "f" s

		set b1 $s
		append b1 [string range $z 2 2]

		set b1 [binary format "H2" $b1]
		set i 3
		set j 4
		}

	# Frequency digit pairs.  Two digits per byte.

	set s [string range $z $i $j]
	set b2 [binary format "H2" $s]

	incr i 2
	incr j 2
	set s [string range $z $i $j]
	set b3 [binary format "H2" $s]

	incr i 2
	incr j 2
	set s [string range $z $i $j]
	set b4 [binary format "H2" $s]

	set freqbytes ""
	append freqbytes $b1 $b2 $b3 $b4

	return $freqbytes
}


###################################################################
# Encode a frequency into Yaesu 3 byte format.
#
# Input:	frequency in MHz
# Returns:	3 byte BCD Yaesu coded frequency
###################################################################
proc Freq2BCD3 { f } \
{
	global GlobalParam
	
	# Note: Yaesu packs two digits per byte, one per nibble.
	# An important exception is the most significant nibble
	# in the most significant byte.  That nibble can be
	# 0-9 or a-f.
	# a-f means 10-15.


	# Check for non-digit chars.
	set rc [regexp {^[0-9\.]*$} $f]
	if {$rc == 0} {set f 0.0}

	if { ($f > $GlobalParam(HighestFreq)) \
		|| ($f < $GlobalParam(LowestFreq)) } \
		{
		set f 0.0
		}

	set z [expr {($f * 1000.0) + .0005}]
	set z [expr {int($z)}]

	if {$f <= 1000.0} \
		{
		set z [format "%06d" $z]

		set s [string range $z 0 1]
		set b1 [binary format "H2" $s]

		set i 2
		set j 3
		} \
	else \
		{
		set z [format "%07d" $z]
		set s [string range $z 0 1]


		regsub -nocase {10} $s "a" s
		regsub -nocase {11} $s "b" s
		regsub -nocase {12} $s "c" s
		regsub -nocase {13} $s "d" s
		regsub -nocase {14} $s "e" s
		regsub -nocase {15} $s "f" s

		set b1 $s
		append b1 [string range $z 2 2]

		set b1 [binary format "H2" $b1]
		set i 3
		set j 4
		}

	# Frequency digit pairs.  Two digits per byte.

	set s [string range $z $i $j]
	set b2 [binary format "H2" $s]

	incr i 2
	incr j 2
	set s [string range $z $i $j]
	set b3 [binary format "H2" $s]

	set freqbytes ""
	append freqbytes $b1 $b2 $b3

	return $freqbytes
}


###################################################################
# Encode a frequency into Yaesu 3 byte plus format.
#
# Input:	frequency in MHz
# Returns:	4 byte BCD Yaesu coded frequency plus flag/mode byte
#
#	The 4th byte returned will be x00, x20, or x40,
#	depending on whether the freq is a 12.5 or 6.25 kHz
#	splinter or not.
#
###################################################################
proc Freq2BCD3p { f } \
{
	global GlobalParam
	
	# Note: Yaesu packs two digits per byte, one per nibble.
	# An important exception is the most significant nibble
	# in the most significant byte.  That nibble can be
	# 0-9 or a-f.
	# a-f means 10-15.


	# Check for non-digit chars.
	set rc [regexp {^[0-9\.]*$} $f]
	if {$rc == 0} {set f 0.0}

	if { ($f > $GlobalParam(HighestFreq)) \
		|| ($f < $GlobalParam(LowestFreq)) } \
		{
		set f 0.0
		}

	# Compensate for roundoff errors.
	set z [expr {($f * 100000.0) + .000005}]
	set z [expr {int($z)}]

	if {$f < 1000.0} \
		{
		set z [format "%08d" $z]

		set s [string range $z 0 1]
		set b1 [binary format "H2" $s]

		set i 2
		set j 3
		} \
	else \
		{
		set z [format "%09d" $z]
		set s [string range $z 0 1]


		regsub -nocase {10} $s "a" s
		regsub -nocase {11} $s "b" s
		regsub -nocase {12} $s "c" s
		regsub -nocase {13} $s "d" s
		regsub -nocase {14} $s "e" s
		regsub -nocase {15} $s "f" s

		set b1 $s
		append b1 [string range $z 2 2]

		set b1 [binary format "H2" $b1]
		set i 3
		set j 4
		}

	# Frequency digit pairs.  Two digits per byte.

	set s [string range $z $i $j]
	set b2 [binary format "H2" $s]

	incr i 2
	incr j 2
	set s [string range $z $i $j]
	set b3 [binary format "H2" $s]

	set flag50 [regexp {50$} $z]
	set flag25 [regexp {25$} $z]

	if {$flag50} \
		{
		# Frequency is a 12.5 kHz splinter
		set b4 [binary format "H2" 40]
		} \
	elseif {$flag25} \
		{
		# Frequency is a 6.25 kHz splinter
		set b4 [binary format "H2" 20]
		} \
	else \
		{
		set b4 [binary format "H2" 00]
		}

	binary scan $b4 "H2" s
#	puts stderr "f= $f  z=$z b4= $s"

	set freqbytes ""
	append freqbytes $b1 $b2 $b3 $b4


	return $freqbytes
}


###################################################################
# Create widgets for the name of this program.
###################################################################
proc MakeTitleFrame { f }\
{
	global DisplayFontSize 
	global Version

	frame $f -relief flat -borderwidth 3

	# set s [format "tk120 v%s" $Version]
	set s [format "tk120"]

	label $f.lab -text $s \
		-background blue \
		-foreground white \
		-relief raised \
		-borderwidth 3 \
		-font $DisplayFontSize 

	set s ""
	append s [format "Version %s\n" $Version]
	append s "Experimental Utility\n"
	append s "for the Yaesu VR-120 Receiver\n"
	append s "Copyright 2001 - 2005, Bob Parnass"

	label $f.use -text $s \
		-background black \
		-foreground white \
		-relief raised \
		-borderwidth 3

	pack $f.lab $f.use -side top -padx 0 -pady 0 \
		-fill y -fill x -expand true

	return $f
}


###################################################################
# Create frame for display parameters. 
###################################################################
proc MakeDisplayFrame { f }\
{

	frame $f -relief groove -borderwidth 3
	label $f.lab -text "Display, Keypad Settings" \
		-borderwidth 3
	pack $f.lab -side top

	frame $f.b -relief flat -borderwidth 3
	MakeDispWidgets  $f.b
	pack $f.b -side top -expand true -fill both

	return $f
}

proc MakeDispWidgets {f} \
{
	global GlobalParam


        label $f.llamp -text "Lamp" -borderwidth 3
	tk_optionMenu $f.lamp GlobalParam(Lamp) OFF ON AUTO

	checkbutton $f.lock -text "Lock Keypad" \
		-variable GlobalParam(Lock) \
		-onvalue 1 -offvalue 0


	checkbutton $f.beep -text "Keys beep when pressed" \
		-variable GlobalParam(Beep) \
		-onvalue 0 -offvalue 1

	checkbutton $f.smbeep -text "S-Meter beeps" \
		-variable GlobalParam(SMeterBeep) \
		-onvalue 1 -offvalue 0

	set hint ""
	append hint "You can configure the radio to emit a tone "
	append hint "whose pitch varies with signal strength.\n"
	append hint "Note: The tone is annoying."
	balloonhelp_for $f.smbeep $hint


	checkbutton $f.smsens -text "Decrease S-Meter sensitivity" \
		-variable GlobalParam(SMeterSens) \
		-onvalue 1 -offvalue 0

	set hint ""
	append hint "You can desensitize the S-Meter.\n"
	append hint "This is useful for making signal "
	append hint "strength comparisons among strong signals."
	balloonhelp_for $f.smsens $hint

	grid $f.llamp -row 6 -column 0 -sticky w
	grid $f.lamp -row 6 -column 1 -sticky e
	grid $f.lock -row 12 -column 0 -sticky w -columnspan 2
	grid $f.beep -row 14 -column 0 -sticky w -columnspan 2
	grid $f.smbeep -row 16 -column 0 -sticky w -columnspan 2
	grid $f.smsens -row 18 -column 0 -sticky w -columnspan 2

	return

}


###################################################################
# Create 10 search banks. 
###################################################################
proc MakeSearchFrame { f }\
{
	global GlobalParam


	frame $f -relief groove -borderwidth 3

	frame $f.rb -relief groove -borderwidth 3
	frame $f.misc -relief sunken -borderwidth 3
	set m $f.misc
	set r $f.rb



	radiobutton $r.vfosearch -text "VFO Search" \
		-variable GlobalParam(LimitSearch) -value 0

	radiobutton $r.limitsearch -text "Limit Search" \
		-variable GlobalParam(LimitSearch) -value 1


	pack $r.vfosearch $r.limitsearch -side left
	pack $r -side top -padx 3 -pady 3


	checkbutton $m.flexstep \
		-text "Align frequency\nwith tuning step" \
		-variable GlobalParam(FlexStep) \
		-onvalue 0 -offvalue 1

	set hint ""
	append hint "This Flex Step feature coerces the frequency to "
	append hint "be a whole multiple of the tuning step size.\n\n"
	append hint "Example: Disable this and set the Tuning Step "
	append hint "size to 25 if you want to tune "
	append hint "a sequence like 460.0125, 460.0375, 460.0625, ..."
	balloonhelp_for $m.flexstep $hint



	label $m.lvfo -text "VFO" -borderwidth 3

	entry $m.vfo -width 10 \
		-textvariable GlobalParam(VFOFreq) \
		-background white 

	label $m.lvfomode -text "VFO Mode" -borderwidth 3
	tk_optionMenu $m.vfomode GlobalParam(VFOMode) \
		AUTO NFM WFM AM

        label $m.ltstep -text "Tuning Step" -borderwidth 3
	tk_optionMenu $m.tstep GlobalParam(TuningStep) \
		AUTO 5 6.25 9 10 12.5 15 20 25 30 50 100


        label $m.lftstep -text "Fast Tuning Step" -borderwidth 3
	tk_optionMenu $m.ftstep GlobalParam(FastTuningStep) \
		10kHz 100kHz 1MHz 10MHz 100MHz


        label $m.lsresume -text "VFO Scan Resume" -borderwidth 3
	tk_optionMenu $m.sresume GlobalParam(ScanResumeVFO) \
		PAUSE BUSY HOLD


	grid $m.flexstep -row 2 -column 0 -sticky w -columnspan 2
	grid $m.lvfo -row 6 -column 0 -sticky w
	grid $m.vfo -row 6 -column 1 -sticky e
	grid $m.lvfomode -row 10 -column 0 -sticky w
	grid $m.vfomode -row 10 -column 1 -sticky e

	grid $m.ltstep -row 15 -column 0 -sticky w
	grid $m.tstep -row 15 -column 1 -sticky ew
	grid $m.lftstep -row 20 -column 0 -sticky w
	grid $m.ftstep -row 20 -column 1 -sticky ew
	grid $m.lsresume -row 25 -column 0 -sticky w
	grid $m.sresume -row 25 -column 1 -sticky ew

	set hint ""
	append hint "The Fast Tuning step is used "
	append hint "when you hold down the FUNC key while "
	append hint "twisting the tuning knob at the same time."
	balloonhelp_for $m.lftstep $hint
	balloonhelp_for $m.ftstep $hint

	pack $m -side top -padx 3 -pady 3


	label $f.lab -text "\nLimit Search Banks" -borderwidth 3
	pack $f.lab -side top -padx 3 -pady 3

	frame $f.b -relief flat -borderwidth 3


	for {set i 0} {$i < 9} {incr i} \
		{
		MakeSearchBank $f.b $i
		}

	pack $f.b -side top -fill both -expand true -padx 3 -pady 3

	return $f
}


###################################################################
# Create frame for tweaking parameters. 
###################################################################
proc MakeTweaksFrame { f }\
{
	global GlobalParam

	frame $f -relief groove -borderwidth 3

	set s "Experimental Tweaks\n"
	append s "(Effects, if any, are unknown.\n"
	append s "Use at your own risk.)"
	label $f.lab -text $s -borderwidth 3

	pack $f.lab -side top

	frame $f.b -relief flat -borderwidth 3
	MakeTweakWidgets  $f.b

	pack $f.b -side top -expand true -fill both

	return $f
}

proc MakeTweakWidgets {f} \
{
	global GlobalParam

	checkbutton $f.userport -text "User Port" \
		-variable GlobalParam(UserPort) \
		-onvalue 1 -offvalue 0


	set hint ""
	append hint "Caution: Experiment Tweaks are undocumented "
	append hint "and their effect (if any) is unknown."
	balloonhelp_for $f $hint
	# balloonhelp_for $f.userport $hint

	# Note: Set Menu Item 26 is a hidden "User P" (User Port) menu!
	# It has choices of OFF and ON, with OFF being the default.

        label $f.lsetmenuitem -text "Set Menu Item" -borderwidth 3
        tk_optionMenu $f.setmenuitem GlobalParam(SetMenuItem) \
                0 1 2 3 4 5 6 7 8 9 \
		10 11 12 13 14 15 16 17 18 19 \
		20 21 22 23 24 25 26

        label $f.lbits -text "\nBit Flags (usually off)" \
		-borderwidth 3

	checkbutton $f.flag0bit0 -text "1A14 Bit 0" \
		-variable GlobalParam(Flag0Bit0) \
		-onvalue 1 -offvalue 0

	checkbutton $f.flag0bit1 -text "1A14 Bit 1" \
		-variable GlobalParam(Flag0Bit1) \
		-onvalue 1 -offvalue 0



	checkbutton $f.flag1bit0 -text "1A15 Bit 0" \
		-variable GlobalParam(Flag1Bit0) \
		-onvalue 1 -offvalue 0

	checkbutton $f.flag1bit6 -text "1A15 Bit 6" \
		-variable GlobalParam(Flag1Bit6) \
		-onvalue 1 -offvalue 0

	checkbutton $f.flag1bit7 -text "1A15 Bit 7" \
		-variable GlobalParam(Flag1Bit7) \
		-onvalue 1 -offvalue 0

        label $f.lx1a11 -text "1A11 (hex)" -borderwidth 3
        entry $f.x1a11 -width 3\
		-textvariable GlobalParam(1A11) -background yellow

	label $f.lx1a0c -text "1A0C (hex)" -borderwidth 3
	entry $f.x1a0c -width 3 \
		-textvariable GlobalParam(1A0C) -background white 

	label $f.lx1a0e -text "1A0E (hex)" -borderwidth 3
	entry $f.x1a0e -width 3 \
		-textvariable GlobalParam(1A0E) -background white 

	label $f.lx1a10 -text "1A10 (hex)" -borderwidth 3
	entry $f.x1a10 -width 3 \
		-textvariable GlobalParam(1A10) -background white 


	grid $f.userport -row 1 -column 0 -sticky w -columnspan 2
	grid $f.lsetmenuitem -row 2 -column 0 -sticky w
	grid $f.setmenuitem -row 2 -column 1 -sticky e

	grid $f.lbits -row 3 -column 0 -sticky w
	grid $f.flag0bit0 -row 4 -column 0 -sticky w -columnspan 2
	grid $f.flag0bit1 -row 6 -column 0 -sticky w -columnspan 2

	grid $f.flag1bit0 -row 8 -column 0 -sticky w -columnspan 2
	grid $f.flag1bit6 -row 12 -column 0 -sticky w -columnspan 2
	grid $f.flag1bit7 -row 14 -column 0 -sticky w -columnspan 2

	grid $f.lx1a0c -row 18 -column 0 -sticky w
	grid $f.x1a0c -row 18 -column 1 -sticky w

	grid $f.lx1a0e -row 20 -column 0 -sticky w
	grid $f.x1a0e -row 20 -column 1 -sticky w

	grid $f.lx1a10 -row 22 -column 0 -sticky w
	grid $f.x1a10 -row 22 -column 1 -sticky w

	grid $f.lx1a11 -row 24 -column 0 -sticky w
	grid $f.x1a11 -row 24 -column 1 -sticky w

	return

}



###################################################################
# Create one a set of widgets for one search bank. 
###################################################################
proc MakeSearchBank { f bn }\
{
	global LimitScan
	global GlobalParam

	if {$bn == 8} \
		{
		label $f.lab$bn -text "L--" -borderwidth 3
		} \
	else \
		{
		label $f.lab$bn -text "L0$bn" -borderwidth 3
		}

	entry $f.lower$bn -width 10 \
		-textvariable LimitScan($bn,lower) \
		-background white 

	tk_optionMenu $f.lmodemenu$bn LimitScan($bn,lmode) \
		AUTO NFM WFM AM

	entry $f.upper$bn -width 10 \
		-textvariable LimitScan($bn,upper) \
		-background white 

	tk_optionMenu $f.umodemenu$bn LimitScan($bn,umode) \
		AUTO NFM WFM AM

	grid $f.lab$bn -row $bn -column 1
	grid $f.lower$bn -row $bn -column 2
	grid $f.lmodemenu$bn -row $bn -column 3 -sticky ew
	grid $f.upper$bn -row $bn -column 4
	grid $f.umodemenu$bn -row $bn -column 5 -sticky ew

	return $f
}


###################################################################
# Create 8 Dual Watch pairs. 
###################################################################
proc MakeDWFrame { f }\
{

	frame $f -relief groove -borderwidth 3

	###########################################
	# Priority or Dual Watch?  frame
	###########################################
	set r $f.r
	frame $r -relief groove -borderwidth 3

	radiobutton $r.dualwatch -text "Dual Watch" \
		-variable GlobalParam(DWorP) -value 0

	radiobutton $r.priority -text "Priority" \
		-variable GlobalParam(DWorP) -value 1

	pack $r.priority $r.dualwatch -side left
	pack $r -side top -padx 3 -pady 3

	###########################################
	# Priority settings frame
	###########################################
	set p $f.p
	frame $p -relief sunken -borderwidth 3


	# Priority settings
	label $p.labp -text "Priority Frequency" -borderwidth 3
	entry $p.priority -width 10 -textvariable Priority \
		-background white 

	label $p.labpmode -text "Priority Mode" -borderwidth 3
	tk_optionMenu $p.prioritymode PriorityMode \
		AUTO NFM WFM AM


	grid $p.labp  -row 3 -column 0 -sticky w
	grid $p.priority -row 3 -column 1 -sticky e

	grid $p.labpmode  -row 4 -column 0 -sticky w
	grid $p.prioritymode -row 4 -column 1 -sticky e
	pack $p -side top -padx 3 -pady 3

####


	###########################################
	# Dual Watch settings frame
	###########################################
	set d $f.d
	frame $d -relief flat -borderwidth 3

	label $d.lab -text "\nDual Watch" -borderwidth 3

        label $d.lwhichdw -text "Dual Watch Bank" -borderwidth 3
        tk_optionMenu $d.whichdw GlobalParam(WhichDW) \
                0 1 2 3 4 5 6 7 8

        label $d.lsresume -text "Dual Watch Scan Resume" -borderwidth 3
	tk_optionMenu $d.sresume GlobalParam(ScanResumeDW) \
		PAUSE BUSY HOLD

	grid $d.lab  -row 4 -column 0 -sticky ew -columnspan 2

	grid $d.lwhichdw  -row 8 -column 0 -sticky w
	grid $d.whichdw -row 8 -column 1 -sticky e
	grid $d.lsresume  -row 12 -column 0 -sticky w
	grid $d.sresume -row 12 -column 1 -sticky e

	pack $d -side top -padx 3 -pady 3


	frame $f.b -relief groove -borderwidth 3


	for {set i 0} {$i < 9} {incr i} \
		{
		MakeDWBank $f.b $i
		}
####
#	###########################################
#	# Which Dual Watch?  frame
#	###########################################
#
#	set w $f.which
#	frame $w -relief flat -borderwidth 3
#
#        label $w.lwhichdw -text "Dual Watch Bank" -borderwidth 3
#        tk_optionMenu $w.whichdw GlobalParam(WhichDW) \
#                0 1 2 3 4 5 6 7 8
#
#
#	grid $w.lwhichdw  -row 4 -column 0 -sticky w
#	grid $w.whichdw -row 4 -column 1 -sticky e
#
#	pack $w -side top
#
#	label $f.lab -text "\nDual Watch" -borderwidth 3
#	pack $f.lab -side top
#
#	###########################################
#	# Dual Watch settings frame
#	###########################################
#	frame $f.b -relief flat -borderwidth 3
#
#
#	for {set i 0} {$i < 9} {incr i} \
#		{
#		MakeDWBank $f.b $i
#		}

	pack $f.b -side top -fill y -expand true -padx 3 -pady 3

	return $f
}

###################################################################
# Create one a set of widgets for one dual watch bank. 
###################################################################
proc MakeDWBank { f bn }\
{
	global DualWatch
	global GlobalParam

	if {$bn == 8} \
		{
		label $f.lab$bn -text "A--" -borderwidth 3
		} \
	else \
		{
		label $f.lab$bn -text "A0$bn" -borderwidth 3
		}

	entry $f.lower$bn -width 10 \
		-textvariable DualWatch($bn,first) \
		-background white 

	tk_optionMenu $f.lmodemenu$bn DualWatch($bn,mode1) \
		AUTO NFM WFM AM

	entry $f.upper$bn -width 10 \
		-textvariable DualWatch($bn,second) \
		-background white 

	tk_optionMenu $f.umodemenu$bn DualWatch($bn,mode2) \
		AUTO NFM WFM AM

	grid $f.lab$bn -row $bn -column 1
	grid $f.lower$bn -row $bn -column 2
	grid $f.lmodemenu$bn -row $bn -sticky ew -column 3
	grid $f.upper$bn -row $bn -column 4
	grid $f.umodemenu$bn -row $bn -sticky ew -column 5

	return $f
}


###################################################################
# Create widgets for 9 Preset Memories. 
###################################################################
proc MakePresetFrame { f }\
{

	frame $f -relief groove -borderwidth 3
	label $f.lab -text "Preset Memories" -borderwidth 3
	pack $f.lab -side top

	frame $f.b -relief flat -borderwidth 3

	for {set i 0} {$i < 11} {incr i} \
		{
		MakePresetBank $f.b $i
		}

	pack $f.b -side top

	set hint ""
	append hint "Preprogrammed frequencies "
	append hint "serve as 'starting points' in Preset mode."
	balloonhelp_for $f.b $hint



	# Weather WX channel preset

	set w $f.w
	frame $w -relief groove -borderwidth 3

	label $w.lwx -text "WX" -borderwidth 3

	entry $w.wx -width 10 \
		-textvariable GlobalParam(WXFreq) \
		-background white 

	tk_optionMenu $w.wxmode GlobalParam(WXMode) \
		AUTO NFM WFM AM

	grid $w.lwx -row 2 -column 0 -sticky w
	grid $w.wx -row 2 -column 1
	grid $w.wxmode -row 2 -column 2 -sticky e
 

	pack $w -side top -padx 3 -pady 3

	set hint ""
	append hint "The weather frequency "
	append hint "is stored in the radio, "
	append hint "but details of its use are unknown."
	balloonhelp_for $w $hint


	return $f
}

###################################################################
# Create one a set of widgets for one Preset bank. 
###################################################################
proc MakePresetBank { f bn }\
{
	global Preset
	global GlobalParam

	label $f.lab$bn -text "$bn" -borderwidth 3

	entry $f.freq$bn -width 10 \
		-textvariable Preset($bn,freq) \
		-background white 

	tk_optionMenu $f.presetmodemenu$bn Preset($bn,mode) \
		AUTO NFM WFM AM


	grid $f.lab$bn -row $bn -column 1
	grid $f.freq$bn -row $bn -column 2
	grid $f.presetmodemenu$bn -row $bn -sticky ew -column 3

	return $f
}



###################################################################
# Create frame for misc parameters. 
###################################################################
proc MakeMiscFrame { f }\
{

	frame $f -relief groove -borderwidth 3
	label $f.lab -text "Misc. Settings" \
		-borderwidth 3
	pack $f.lab -side top

	frame $f.b -relief flat -borderwidth 3
	MakeMiscWidgets  $f.b
	pack $f.b -side top -expand true -fill both

	return $f
}


###################################################################
# Create widgets for misc. parameters. 
###################################################################
proc MakeMiscWidgets { f } \
{
	global GlobalParam
	global Priority
	global PriorityMode

        label $f.labbattery -text "Battery Save Interval" -borderwidth 3
        tk_optionMenu $f.battery GlobalParam(BatterySaver) \
                OFF 1 3 5 7 9

        label $f.labresume -text "Scan Resume (sec.)" -borderwidth 3
        tk_optionMenu $f.resume GlobalParam(Resume) \
                1 2 3 4 5 6 7 8 9 10 11 12

        # label $f.lantam -text "AM BCB Antenna" -borderwidth 3

	checkbutton $f.atten -text "Attenuator" \
		-variable GlobalParam(Attenuator) \
		-onvalue 1 -offvalue 0

	checkbutton $f.antam -text "Internal Bar AM BCB Antenna" \
		-variable GlobalParam(AntAM) \
		-onvalue 1 -offvalue 0

	checkbutton $f.antfm -text "Earphone SW/VHF/UHF Antenna" \
		-variable GlobalParam(AntFM) \
		-onvalue 1 -offvalue 0

        label $f.lfcounter -text "Freq counter width (+/-MHz)" \
		-borderwidth 3
	tk_optionMenu $f.fcounter GlobalParam(FCounter) \
		5 10 50 100


        grid $f.labbattery  -row 5 -column 0 -sticky w
        grid $f.battery -row 5 -column 1 -sticky e

        grid $f.labresume  -row 6 -column 0 -sticky w
        grid $f.resume -row 6 -column 1 -sticky e

	grid $f.lfcounter -row 8 -column 0 -sticky w
	grid $f.fcounter -row 8 -column 1 -sticky ew

        grid $f.atten  -row 12 -column 0 -sticky w -columnspan 2
        grid $f.antam  -row 14 -column 0 -sticky w -columnspan 2
        grid $f.antfm -row 16 -column 0 -sticky w -columnspan 2


	return $f
}


###################################################################
# Create frame for Smart Search. 
###################################################################
proc MakeSmartSearchFrame { f }\
{

	frame $f -relief groove -borderwidth 3
	label $f.lab -text "Smart Search\nFrequencies" -borderwidth 3
	pack $f.lab -side top

	frame $f.b -relief flat -borderwidth 3
	MakeSmartSearchWidgets  $f.b

	pack $f.b -side top -expand true -fill y

	return $f
}

###################################################################
# Create widgets for Smart Search. 
###################################################################
proc MakeSmartSearchWidgets { f } \
{
	label $f.labl -text "Lower" -borderwidth 3
	label $f.labu -text "Upper" -borderwidth 3
	label $f.labs -text "Start" -borderwidth 3

	entry $f.lower -width 10 \
		-textvariable SmartSearch(lower) \
		-background white 

	entry $f.upper -width 10 \
		-textvariable SmartSearch(upper) \
		-background white 

	entry $f.start -width 10 \
		-textvariable SmartSearch(start) \
		-background white 

	checkbutton $f.continue -text "Continue" \
		-variable GlobalParam(SSContinue) \
		-onvalue 1 -offvalue 0


	set hint ""
	append hint "Selecting Continue will cause a Smart Search "
	append hint "operation to continue sweeping until all "
	append hint "the Smart Search memories have been filled.\n"
	append hint "Otherwise, the Smart Search will halt after "
	append hint "single sweep."
	balloonhelp_for $f.continue $hint


	grid $f.labl  -row 0 -column 0 -sticky w
	grid $f.lower -row 0 -column 1 -sticky e
	grid $f.labu  -row 1 -column 0 -sticky w
	grid $f.upper -row 1 -column 1 -sticky e
	grid $f.labs  -row 2 -column 0 -sticky w
	grid $f.start -row 2 -column 1 -sticky e
	grid $f.continue -row 4 -column 0 -sticky w -columnspan 2

	return $f
}


###################################################################
# Create frame for Communications parameters. 
###################################################################
proc MakeCommFrame { f }\
{

	frame $f -relief groove -borderwidth 3
	label $f.lab -text "Serial Communication" \
		-borderwidth 3
	pack $f.lab -side top

	frame $f.b -relief flat -borderwidth 3
	MakeCommWidgets  $f.b

	set hint ""
	append hint "Serial Communications fields "
	append hint "are useful for testing tk120. "
	balloonhelp_for $f $hint


	pack $f.b -side top -expand true -fill y

	return $f
}


###################################################################
# Create widgets for Communications params. 
###################################################################
proc MakeCommWidgets { f } \
{
	global GlobalParam

	label $f.labpre -text "Preamble" -borderwidth 3
	entry $f.pre -width 26 \
		-textvariable GlobalParam(PreambleRead) \
		-background yellow 


	label $f.labpost -text "Postamble" -borderwidth 3
	entry $f.post -width 26 \
		-textvariable GlobalParam(PostambleRead) \
		-background yellow 


	label $f.labnmsgs -text "Messages Read" -borderwidth 3
	entry $f.nmsgs -width 5 \
		-textvariable GlobalParam(NmsgsRead) \
		-background yellow 

	label $f.labwadjust -text "Write adjustment" -borderwidth 3
	tk_optionMenu $f.wadjust GlobalParam(Wadjust) \
		0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 \
		20 21 22 23 24 25 26 27 28 29 30

	checkbutton $f.bypassall -text "Bypass All Encoding" \
		-variable GlobalParam(BypassAllEncoding) \
		-onvalue 1 -offvalue 0



	grid $f.labpre  -row 0 -column 0 -sticky w
	grid $f.pre	-row 0 -column 1 -sticky e
	grid $f.labpost  -row 1 -column 0 -sticky w
	grid $f.post	-row 1 -column 1 -sticky e
	grid $f.labnmsgs -row 2 -column 0 -sticky w
	grid $f.nmsgs	-row 2 -column 1 -sticky e
	grid $f.labwadjust	-row 3 -column 0 -sticky w
	grid $f.wadjust	-row 3 -column 1 -sticky e
	grid $f.bypassall	-row 4 -column 0 -columnspan 2

	return $f
}


###################################################################
# Create widgets for Auto Bands. 
###################################################################
proc MakeAutoBandFrame { f }\
{
	global GlobalParam

	frame $f -relief groove -borderwidth 3

	label $f.lab -text "Auto Band (Band Plan)" -borderwidth 3
	pack $f.lab -side top

	ScrollformCreate $f.b
	pack $f.b -expand yes -fill both

	set w [ScrollFormInterior $f.b]

	for {set i 1} {$i < 40} {incr i} \
		{
		MakeAutoBand $w $i
		}


	set hint ""
	set hint [append hint "You can define default mode "]
	set hint [append hint "and step sizes for different "]
	set hint [append hint "sections of the frequency spectrum."]
	balloonhelp_for $f.b $hint


	pack $f.b -side top

	return $f
}

###################################################################
# Create one a set of widgets for one dual watch bank. 
###################################################################
proc MakeAutoBand { f bn }\
{
	global AutoBand
	global GlobalParam

	label $f.lab$bn -text "$bn" -borderwidth 3

	entry $f.freq$bn -width 10 \
		-textvariable AutoBand($bn,freq) \
		-background white 

	tk_optionMenu $f.modemenu$bn AutoBand($bn,mode) \
		NFM WFM AM

	tk_optionMenu $f.stepmenu$bn AutoBand($bn,step) \
		5 6.25 9 10 12.5 15 20 25 30 50 100

	grid $f.lab$bn -row $bn -column 1
	grid $f.freq$bn -row $bn -column 2
	grid $f.modemenu$bn -row $bn -sticky ew -column 3
	grid $f.stepmenu$bn -row $bn -sticky ew -column 4

	return $f
}



###################################################################
# Create widgets for Linked Scan Banks. 
###################################################################
proc MakeLinkedScanFrame { f }\
{

	global GlobalParam

	frame $f -relief groove -borderwidth 3

	label $f.lab -text "Linked Scan Banks" -borderwidth 3
	pack $f.lab -side top

	for {set i 0} {$i < 10} {incr i} \
		{
		MakeBankLink $f $i
		}

	return $f
}

proc MakeBankLink {w bn} \
{
	global BankLink

	checkbutton $w.link$bn -text "$bn" \
		-variable BankLink($bn) \
		-onvalue 1 \
		-offvalue 0

	pack $w.link$bn -side top -anchor center
	return
}

###################################################################
# Create widgets for Skip frequencies and modes. 
###################################################################
proc MakeSkipFrame { f }\
{

	frame $f -relief groove -borderwidth 3
	label $f.lab -text "Skip Frequencies" \
		-borderwidth 3

	button $f.clear -text "Clear All" -command {ClearSkip}


	pack $f.lab -side top
	pack $f.lab $f.clear -side top -padx 3 -pady 3

	ScrollformCreate $f.b
	pack $f.b -expand yes -fill both

	set w [ScrollFormInterior $f.b]

	for {set i 0} {$i < 64} {incr i} \
		{
		MakeSkip $w $i
		}



	set hint ""
	append hint "You can designate frequencies "
	append hint "to skip over during VFO and limit "
	append hint "searches. "
	append hint "This is useful for ignoring "
	append hint "paging, birdy, and other boring "
	append hint "frequencies. "
	balloonhelp_for $f.b $hint

	pack $f.b -side top -padx 3 -pady 3

	return $f
}

###################################################################
# Create one a set of widgets for one skip freq/mode pair. 
###################################################################
proc MakeSkip { f bn }\
{
	global Skip
	global GlobalParam

	label $f.lab$bn -text "S$bn" -borderwidth 3

	entry $f.freq$bn -width 10 \
		-textvariable Skip($bn,freq) \
		-background white 

	tk_optionMenu $f.modemenu$bn Skip($bn,mode) \
		AUTO NFM WFM AM


	grid $f.lab$bn -row $bn -column 0
	grid $f.freq$bn -row $bn -column 1
	grid $f.modemenu$bn -row $bn -sticky ew -column 2

	return $f
}

proc ClearSkip {} \
{
	global Skip

	for {set i 0} {$i < 64} {incr i} \
		{
		set Skip($i,freq) 0.00000
		set Skip($i,mode) NFM
		}
	return
}



###################################################################
# Create widgets for band edge frequencies and modes. 
###################################################################
proc MakeBandEdgeFrame { f }\
{

	frame $f -relief groove -borderwidth 3

	set msg "Band Edges"
	label $f.lab -text $msg \
		-borderwidth 3

	pack $f.lab -side top

	frame $f.b -relief flat -borderwidth 3
	pack $f.b -side top -padx 3 -pady 3


	set hint ""
	append hint "The radio employs bandstacking, "
	append hint "i.e., it remembers the last VFO "
	append hint "frequency used in each band.\n"
	append hint "Band Edge frequencies delimit\n"
	append hint "the bands used with the\n"
	append hint "bandstacking memories."
	balloonhelp_for $f.b $hint

	for {set i 0} {$i < 12} {incr i} \
		{
		MakeBandEdge $f.b $i
		}



	pack $f.b -side top

	return $f
}

###################################################################
# Create one a set of widgets for one band edge freq. 
###################################################################
proc MakeBandEdge { f bn }\
{
	global BandEdge
	global GlobalParam

	label $f.lab$bn -text "$bn" -borderwidth 3

	entry $f.freq$bn -width 10 \
		-textvariable BandEdge($bn,freq) \
		-background white 


	grid $f.lab$bn -row $bn -column 0
	grid $f.freq$bn -row $bn -column 1

	return $f
}


###################################################################
# Create widgets for BandStack Memories. 
###################################################################
proc MakeBandStackFrame { f }\
{

	frame $f -relief groove -borderwidth 3
	label $f.lab -text "Bandstacking Memories" -borderwidth 3
	pack $f.lab -side top

	frame $f.b -relief flat -borderwidth 3

	for {set i 0} {$i < 12} {incr i} \
		{
		MakeBandStackBank $f.b $i
		}

	pack $f.b -side top

	set hint ""
	append hint "Bandstacking memories "
	append hint "remember the last VFO frequency and mode "
	append hint "setting for each band."
	balloonhelp_for $f.b $hint

	return $f
}


###################################################################
# Create one a set of widgets for one BandStack bank. 
###################################################################
proc MakeBandStackBank { f bn }\
{
	global BandStack
	global GlobalParam

	label $f.lab$bn -text "$bn" -borderwidth 3

	entry $f.freq$bn -width 10 \
		-textvariable BandStack($bn,freq) \
		-background white 

	tk_optionMenu $f.bsmodemenu$bn BandStack($bn,mode) \
		AUTO NFM WFM AM


	grid $f.lab$bn -row $bn -column 1
	grid $f.freq$bn -row $bn -column 2
	grid $f.bsmodemenu$bn -row $bn -sticky ew -column 3

	return $f
}


###################################################################
# Encode the information from the data structures into
# the memory image string which can be written to the VR-120.
#
# We don't understand the meaning of all the bytes in
# the memory image.  Therefore, the
# image string must already exist and we will only
# change the bytes which we understand.
#
###################################################################
proc EncodeImage { } \
{
	global GlobalParam
	global Mimage

	if {$GlobalParam(BypassAllEncoding)} \
		{
		puts stderr "EncodeImage: skip encoding"
		return 0
		}

	# puts stderr "EncodeImage: encoding"
	if { ([info exists Mimage] == 0) } \
		{
		puts stderr "EncodeImage: image does not exist"
		return error
		}

	set image $Mimage

	set image [EncodeMisc $image]
	if { [string length $image] == 0} {return error}
	update idletasks

	set image [EncodeTweaks $image]
	if { [string length $image] == 0} {return error}
	update idletasks

	set image [EncodeSmartSearch $image]
	if { [string length $image] == 0} {return error}
	update idletasks

	set image [EncodeScanResume $image]
	if { [string length $image] == 0} {return error}
	update idletasks

	set image [EncodeMemories $image]
	if { [string length $image] == 0} {return error}
	update idletasks

	set image [EncodeSearchBanks $image]
	if { [string length $image] == 0} {return error}
	update idletasks

	set image [EncodeBankLinks $image]
	if { [string length $image] == 0} {return error}
	update idletasks

	set image [EncodeSkip $image]
	if { [string length $image] == 0} {return error}
	update idletasks

	set image [EncodeBandEdges $image]
	if { [string length $image] == 0} {return error}
	update idletasks

	set image [EncodeBandStack $image]
	if { [string length $image] == 0} {return error}
	update idletasks

	set image [EncodeDWBanks $image]
	if { [string length $image] == 0} {return error}
	update idletasks

	set image [EncodePresets $image]
	if { [string length $image] == 0} {return error}
	update idletasks

	set image [EncodeAutoBand $image]
	if { [string length $image] == 0} {return error}
	update idletasks


	# reparse flag byte in tweak section
	# because the value of this field depends on the
	# value of other controls

	DecodeFlagByte3 $image

	set Mimage $image


	return 0
}



###################################################################
# Encode misc
# information into a memory image.
###################################################################
proc EncodeMisc { image } \
{
	global BatterySaver
	global Counter
	global Fstep
	global ImageAddr
	global GlobalParam
	global Mimage
	global Mode
	global Priority
	global PriorityMode
	global Step


	# Priority frequency and mode.

	set s [FreqMode2BCD $Priority $PriorityMode ""]

	scan $ImageAddr(Priority) "%x" first
	set last [expr {$first + 3}]
	set image [string replace $image $first $last $s]


	# VFO frequency

	set s [Freq2BCD3p $GlobalParam(VFOFreq)]
	scan $ImageAddr(VFOFreq) "%x" first
	set last [expr {$first + 3}]
	set image [string replace $image $first $last $s]

	# VFO mode.
	scan $ImageAddr(VFOMode) "%x" first
	set last $first

	set m $GlobalParam(VFOMode)
	if { [info exists Mode($m)] } \
		{
		set c $Mode($m)
		} \
	else \
		{
		set c $Mode(NFM)
		}

	set s [format "%02x" $c]
	set b [binary format "H2" $s]
	set image [string replace $image $first $first $b]


	# Set Attenuator bit
	scan $ImageAddr(FlagByte3) "%x" first
	set byte [string index $image $first]
	set newbyte [AssignBit $byte 5 $GlobalParam(Attenuator)]
	set image [string replace $image $first $first $newbyte]

	# Set AM antenna bit
	scan $ImageAddr(FlagByte0) "%x" first
	set byte [string index $image $first]
	set newbyte [AssignBit $byte 7 $GlobalParam(AntAM)]
	set image [string replace $image $first $first $newbyte]


	# Set FM antenna bit
	scan $ImageAddr(FlagByte1) "%x" first
	set byte [string index $image $first]
	set newbyte [AssignBit $byte 5 $GlobalParam(AntFM)]
	set image [string replace $image $first $first $newbyte]


	# Set Keypad Lock bit
	scan $ImageAddr(FlagByte3) "%x" first
	set byte [string index $image $first]
	set newbyte [AssignBit $byte 1 $GlobalParam(Lock)]
	set image [string replace $image $first $first $newbyte]


	# encode Scan Resume.
	# Scan Resume is stored in one hex byte as one
	# less than what the user sees on the radio's display.
	# Example: Resume of 12 seconds is stored as 0B.

	scan $ImageAddr(Resume) "%x" first
	set c [expr {$GlobalParam(Resume) - 1}]
	set s [format "%02x" $c]
	set b [binary format "H2" $s]
	set image [string replace $image $first $first $b]


	# encode Battery saver
	scan $ImageAddr(BatterySaver) "%x" first
	set c $GlobalParam(BatterySaver)
	if { [info exists BatterySaver($c)] } \
		{
		set c $BatterySaver($c)
		} \
	else \
		{
		set c 1
		}
	set s [format "%02x" $c]
	set b [binary format "H2" $s]
	set image [string replace $image $first $first $b]


	# Tuning Step is stored in one byte.
	# Important note:
	# There is a correlation between the Tuning Step byte
	# and bit 7 of FlagByte3.
	# Bit 7 = 0 if step is AUTO,
	# othewise Bit 7 = 1.

	scan $ImageAddr(VFOStep) "%x" first
	set c $GlobalParam(TuningStep)
	# Translate to hex equivalent
	set c $Step($c)
	set b [binary format "H2" $c]
	set image [string replace $image $first $first $b]

	scan $ImageAddr(FlagByte3) "%x" first
	set b [string range $image $first $first]
	if {$GlobalParam(TuningStep) == "AUTO"} \
		{
		set b [ClearBit $b 7]
		} \
	else \
		{
		set b [SetBit $b 7]
		}
	set image [string replace $image $first $first $b]

	# Bank scan flag
	scan $ImageAddr(FlagByte0) "%x" first
	set b [string range $image $first $first]
	set b [AssignBit $b 4 $GlobalParam(BankScan)]
	set image [string replace $image $first $first $b]

	# Beep tone flag
	scan $ImageAddr(FlagByte0) "%x" first
	set b [string range $image $first $first]
	set b [AssignBit $b 5 $GlobalParam(Beep)]
	set image [string replace $image $first $first $b]

	# encode S-Meter beep alarm flag
	scan $ImageAddr(FlagByte1) "%x" first
	set b [string range $image $first $first]
	set b [AssignBit $b 2 $GlobalParam(SMeterBeep)]
	set image [string replace $image $first $first $b]

	# encode Decrease S-Meter sensitivity flag
	scan $ImageAddr(FlagByte1) "%x" first
	set b [string range $image $first $first]
	set b [AssignBit $b 1 $GlobalParam(SMeterSens)]
	set image [string replace $image $first $first $b]


	# encode Freq Counter
	scan $ImageAddr(FCounter) "%x" first
	set t $GlobalParam(FCounter)
	if { [info exists Counter($t)] } \
		{
		set byte $Counter($t)
		} \
	else \
		{
		set byte $Counter(50)
		}
	set newbyte [binary format "H2" $byte ]
	set image [string replace $image $first $first $newbyte]


	# encode 2 Lamp bits
	scan $ImageAddr(Lamp) "%x" first
	set byte [string index $image $first]

	if {$GlobalParam(Lamp) == "AUTO"} \
		{
		set byte [ClearBit $byte 6]
		set byte [ClearBit $byte 7]
		} \
	elseif {$GlobalParam(Lamp) == "ON"} \
		{
		set byte [ClearBit $byte 6]
		set byte [SetBit $byte 7]
		} \
	else \
		{
		set byte [SetBit $byte 6]
		set byte [ClearBit $byte 7]
		}
	set newbyte $byte
	set image [string replace $image $first $first $newbyte]


	# Set VFO/Limit Search bit
	scan $ImageAddr(FlagByte0) "%x" first
	set byte [string index $image $first]
	set newbyte [AssignBit $byte 3 $GlobalParam(LimitSearch)]
	set image [string replace $image $first $first $newbyte]

	# Fast Tuning Step is stored in one byte
	scan $ImageAddr(FastTuningStep) "%x" first
	set s $GlobalParam(FastTuningStep)
	set s [format "%02x" $Fstep($s)]
	set b [binary format "H2" $s]
	set image [string replace $image $first $first $b]


	# Set DW or Priority bit
	scan $ImageAddr(FlagByte0) "%x" first
	set byte [string index $image $first]
	set newbyte [AssignBit $byte 2 $GlobalParam(DWorP)]
	set image [string replace $image $first $first $newbyte]


	# Set Flex Step bit
	scan $ImageAddr(FlagByte0) "%x" first
	set byte [string index $image $first]
	set newbyte [AssignBit $byte 6 $GlobalParam(FlexStep)]
	set image [string replace $image $first $first $newbyte]


	# encode Which DW Bank.
	# Dual Watch bank number is stored in one byte.
	scan $ImageAddr(WhichDW) "%x" first
	set c $GlobalParam(WhichDW)
	set s [format "%02x" $c]
	set b [binary format "H2" $s]
	set image [string replace $image $first $first $b]

	return $image
}


###################################################################
# Encode tweak information into a memory image.
###################################################################
proc EncodeTweaks { image } \
{
	global ImageAddr
	global GlobalParam
	global Mimage

	# Set Menu Item is stored in one byte
	scan $ImageAddr(SetMenuItem) "%x" first
	set c [expr {$GlobalParam(SetMenuItem) - 0}]
	set s [format "%02x" $c]
	set b [binary format "H2" $s]
	set image [string replace $image $first $first $b]

	# encode tweak flag
	scan $ImageAddr(FlagByte0) "%x" first
	set b [string range $image $first $first]
	set b [AssignBit $b 0 $GlobalParam(Flag0Bit0)]
	set image [string replace $image $first $first $b]

	# encode tweak flag
	scan $ImageAddr(FlagByte0) "%x" first
	set b [string range $image $first $first]
	set b [AssignBit $b 1 $GlobalParam(Flag0Bit1)]
	set image [string replace $image $first $first $b]

	# encode User Port tweak flag
	scan $ImageAddr(FlagByte1) "%x" first
	set b [string range $image $first $first]
	set b [AssignBit $b 3 $GlobalParam(UserPort)]
	set image [string replace $image $first $first $b]

	# encode tweak flag
	scan $ImageAddr(FlagByte1) "%x" first
	set b [string range $image $first $first]
	set b [AssignBit $b 0 $GlobalParam(Flag1Bit0)]
	set image [string replace $image $first $first $b]

	# encode tweak flag
	scan $ImageAddr(FlagByte1) "%x" first
	set b [string range $image $first $first]
	set b [AssignBit $b 6 $GlobalParam(Flag1Bit6)]
	set image [string replace $image $first $first $b]


	# encode tweak flag
	scan $ImageAddr(FlagByte1) "%x" first
	set b [string range $image $first $first]
	set b [AssignBit $b 7 $GlobalParam(Flag1Bit7)]
	set image [string replace $image $first $first $b]

#	scan $ImageAddr(FlagByte3) "%x" first
#	set b [binary format "H2" $GlobalParam(1A11)]
#	set image [string replace $image $first $first $b]

	scan $ImageAddr(1A0C) "%x" first
	set b [binary format "H2" $GlobalParam(1A0C)]
	set image [string replace $image $first $first $b]

	scan $ImageAddr(1A0E) "%x" first
	set b [binary format "H2" $GlobalParam(1A0E)]
	set image [string replace $image $first $first $b]

	scan $ImageAddr(1A10) "%x" first
	set b [binary format "H2" $GlobalParam(1A10)]
	set image [string replace $image $first $first $b]


	return $image
}

###################################################################
# Encode Smart Search lower, upper, and starting frequencies
# into a memory image.
###################################################################
proc EncodeSmartSearch { image } \
{
	global GlobalParam
	global ImageAddr
	global Mimage
	global SmartSearch


	# Smart Search frequency, lower
	set s [Freq2BCD3p $SmartSearch(lower)]

	scan $ImageAddr(SmartSearchL) "%x" first
	set last [expr {$first + 3}]
	set image [string replace $image $first $last $s]

	# Smart Search frequency, upper
	set s [Freq2BCD3p $SmartSearch(upper)]
	scan $ImageAddr(SmartSearchU) "%x" first
	set last [expr {$first + 3}]
	set image [string replace $image $first $last $s]

	# Smart Search frequency, starting freq
	set s [Freq2BCD3p $SmartSearch(start)]
	scan $ImageAddr(SmartSearchS) "%x" first
	set last [expr {$first + 3}]
	set image [string replace $image $first $last $s]

	# Set Single/Continue bit
	scan $ImageAddr(FlagByte1) "%x" first
	set byte [string index $image $first]
	set newbyte [AssignBit $byte 4 $GlobalParam(SSContinue)]
	set image [string replace $image $first $first $newbyte]

	return $image
}

###################################################################
# Encode the memory channel frequency, mode, and preferential
# flag information into a memory image.
###################################################################

proc EncodeMemories { image } \
{
	global HideByte
	global MemFreq
	global MemLabel
	global MemMode
	global MemPref

	global Mimage
	global ImageAddr
	global Mode

	# Frequency.

	scan $ImageAddr(MemoryFreqs) "%x" first

	for {set bn 0} {$bn < 10} {incr bn} \
		{
		set ch [expr {$bn * 100}]
		for {set i 0} {$i < 64} {incr i} \
			{

			set mode [string toupper $MemMode($ch)]
			set b [ FreqMode2BCD $MemFreq($ch) \
				$mode $MemPref($ch) ]

			set last [expr {$first + 3}]
			set image [string replace $image $first \
				$last $b]

			incr ch
			incr first 4
			}
		}


	# Label.

	scan $ImageAddr(MemoryLabels) "%x" first

	for {set bn 0} {$bn < 10} {incr bn} \
		{
		set ch [expr {$bn * 100}]
		for {set i 0} {$i < 64} {incr i} \
			{
			set s [format "%-6s" $MemLabel($ch)]
			set s [string range $s 0 5]

			# VR-120 labels must be upper case.
			set s [string toupper $s]

			set last [expr {$first + 5}]
			set image [string replace $image $first $last $s]

			incr ch
			incr first 6
			}
		}

	return $image
}


###################################################################
# Encode the Skip channel frequency, and mode
# information into a memory image.
###################################################################

proc EncodeSkip { image } \
{
	global Skip
	global Mimage
	global ImageAddr

	# Skip Frequency.

	scan $ImageAddr(SkipFreqs) "%x" first
	set last [expr {$first + 3}]

	for {set ch 0} {$ch < 64} {incr ch} \
		{
		set mode [string toupper $Skip($ch,mode)]

		set s [FreqMode2BCD $Skip($ch,freq) $mode "" ]
		set image [string replace $image $first $last $s]

		incr first 4
		incr last 4
		}


	return $image
}


###################################################################
# Encode the Band Edge frequency
# into a memory image.
###################################################################

proc EncodeBandEdges { image } \
{
	global BandEdge
	global Mimage
	global ImageAddr

	# Band Edge Frequency.

	scan $ImageAddr(BandEdgeFreq) "%x" first
	set last [expr {$first + 3}]

	set mode AUTO
	for {set ch 0} {$ch < 12} {incr ch} \
		{
		# set mode [string toupper $BandEdge($ch,mode)]

		set s [FreqMode2BCD $BandEdge($ch,freq) $mode "" ]
		set image [string replace $image $first $last $s]

		incr first 4
		incr last 4
		}


	return $image
}


###################################################################
# Encode the Preset Memory channel frequency and mode
# information into a memory image.
###################################################################

proc EncodeBandStack { image } \
{
	global GlobalParam
	global ImageAddr
	global Mode
	global BandStack


	# Frequency and mode

	scan $ImageAddr(BandStackFreq) "%x" first

	for {set ch 0} {$ch < 12} {incr ch} \
		{

		set s [FreqMode2BCD $BandStack($ch,freq) \
			$BandStack($ch,mode) "" ]

		set last [expr {$first + 3}]
		set image [string replace $image $first $last $s]

		incr first 4
		}

	return $image
}


###################################################################
# Encode the Preset Memory channel frequency and mode
# information into a memory image.
###################################################################

proc EncodePresets { image } \
{
	global GlobalParam
	global ImageAddr
	global Mode
	global Preset


	# Frequency and mode

	scan $ImageAddr(PresetFreqs) "%x" first

	for {set ch 0} {$ch < 11} {incr ch} \
		{

		set s [FreqMode2BCD $Preset($ch,freq) \
			$Preset($ch,mode) "" ]

		set last [expr {$first + 3}]
		set image [string replace $image $first $last $s]

		incr first 4
		}

####


	# WX frequency

	set s [Freq2BCD3p $GlobalParam(WXFreq)]
	scan $ImageAddr(WXFreq) "%x" first
	set last [expr {$first + 3}]
	set image [string replace $image $first $last $s]

	# WX mode.
	scan $ImageAddr(WXMode) "%x" first
	set last $first

	set m $GlobalParam(WXMode)
	if { [info exists Mode($m)] } \
		{
		set c $Mode($m)
		} \
	else \
		{
		set c $Mode(NFM)
		}

	set s [format "%02x" $c]
	set b [binary format "H2" $s]
	set image [string replace $image $first $first $b]

####


	return $image
}


###################################################################
# Encode the limit search bank frequencies and modes
# information into a memory image.
###################################################################

proc EncodeSearchBanks { image } \
{
	global ImageAddr
	global LimitScan
	global Mode

	# First frequency.

	scan $ImageAddr(SearchFreqFirst) "%x" first

	for {set ch 0} {$ch < 9} {incr ch} \
		{
		if {$ch == 8} \
			{
			scan $ImageAddr(SearchFreqFirstD) "%x" first
			}
		set s [FreqMode2BCD $LimitScan($ch,lower) \
			$LimitScan($ch,lmode) "" ]
		set last [expr {$first + 3}]
		set image [string replace $image $first $last $s]
		incr first 4
		}


	# Second frequency.

	scan $ImageAddr(SearchFreqSecond) "%x" first

	for {set ch 0} {$ch < 9} {incr ch} \
		{
		if {$ch == 8} \
			{
			scan $ImageAddr(SearchFreqSecondD) "%x" first
			}
		set s [FreqMode2BCD $LimitScan($ch,upper) \
			$LimitScan($ch,umode) "" ]
		set last [expr {$first + 3}]
		set image [string replace $image $first $last $s]
		incr first 4
		}


	return $image
}


###################################################################
# Encode the limit search bank frequencies and modes
# information into a memory image.
###################################################################

proc EncodeBankLinks { image } \
{
	global ImageAddr
	global BankLink


	# Bank links.

	scan $ImageAddr(BankLinks) "%x" first

	# Banks 0 - 7.

	set byte [binary format "H2" 00]
	for {set bn 0} {$bn < 8} {incr bn} \
		{
		set i [expr {7 - $bn}]
		if {$BankLink($bn) > 0} \
			{
			set byte [SetBit $byte $i]
			}
		}

	set image [string replace $image $first $first $byte]

	# Banks 8 - 9 flags go in the next consecutive byte.

	incr first
	set byte [binary format "H2" 00]
	for {set bn 8} {$bn < 10} {incr bn} \
		{
		set i [expr {7 - ($bn - 8)}]
		if {$BankLink($bn) > 0} \
			{
			set byte [SetBit $byte $i]
			}
		}

	set image [string replace $image $first $first $byte]

	return $image
}


###################################################################
# Encode the Dual Watch bank frequencies and mode
# information into a memory image.
###################################################################

proc EncodeDWBanks { image } \
{
	global ImageAddr
	global DualWatch
	global Mode


	# First frequency.

	scan $ImageAddr(DualWatchFirst) "%x" first

	for {set ch 0} {$ch < 9} {incr ch} \
		{
		if {$ch == 8} \
			{
			scan $ImageAddr(DualWatchFirstD) "%x" first
			}
		set mode [string toupper $DualWatch($ch,mode1)]
		set s [FreqMode2BCD $DualWatch($ch,first) $mode "" ]
		set last [expr {$first + 3}]
		set image [string replace $image $first $last $s]
		incr first 4
		}


	# Second frequency.

	scan $ImageAddr(DualWatchSecond) "%x" first

	for {set ch 0} {$ch < 9} {incr ch} \
		{
		if {$ch == 8} \
			{
			scan $ImageAddr(DualWatchSecondD) "%x" first
			}
		set mode [string toupper $DualWatch($ch,mode2)]
		set s [FreqMode2BCD $DualWatch($ch,second) $mode "" ]
		set last [expr {$first + 3}]
		set image [string replace $image $first $last $s]
		incr first 4
		}

	return $image
}


###################################################################
# Encode the Auto Band frequency, mode, and step
# information into a memory image.
###################################################################

proc EncodeAutoBand { image } \
{
	global AutoBand
	global AutoBandMode
	global RAutoBandMode

	global ImageAddr
	global Mode

	# Frequency.

	scan $ImageAddr(AutoBand) "%x" first

	for {set ch 1} {$ch < 40} {incr ch} \
		{
		set s [Freq2BCD3 $AutoBand($ch,freq)]
		set last [expr {$first + 2}]
		set image [string replace $image $first $last $s]

		incr first 4
		}

	# Step/mode byte
	scan $ImageAddr(AutoBandSM) "%x" first

	for {set ch 1} {$ch < 40} {incr ch} \
		{
		set j $ch
		incr j -1
		set mode $AutoBand($j,mode)
		set step $AutoBand($j,step)


		if { [info exists RAutoBandMode($mode,$step)] } \
			{
			set s $RAutoBandMode($mode,$step)
			} \
		else \
			{
			set s $RAutoBandMode(NFM,5)
			}

		set b [binary format "H2" $s]
		set image [string replace $image $first $first $b]
		incr first 4
		}
	return $image
}


###################################################################
# Pop up a window which says "Please wait..."
###################################################################
proc MakeWait { } \
{
	global DisplayFontSize


	toplevel .wait

	set w .wait
	wm title $w "tk120 running"

	label $w.lab -font $DisplayFontSize -text "Please wait ..."

	pack $w.lab

	update idletasks
	waiter 500
	return
}

###################################################################
# Kill the window which says "Please wait..."
###################################################################
proc KillWait { } \
{
	catch {destroy .wait}
	update idletasks
}

###################################################################
# ValidateData tests the data.
# It pops up a window with error and/or warning messages.
# If there are warnings but no errors, the user can elect
# to continue or cancel the current operation.
#
# Returns:
#	0	- continue
#	1	- cancel the current operation
###################################################################
proc ValidateData { } \
{
	global AutoBand
	global BandEdge
	global BandStack
	global DualWatch
	global Emsg
	global GlobalParam
	global MemFreq
	global MemMode
	global LimitScan
	global Preset
	global Priority
	global PriorityMode
	global Skip
	global SmartSearch

	if { [info exists MemFreq(0)] == 0 } \
		{
		# No data to validate.
		return 1
		}


	set Emsg ""
	set nerror 0
	set nwarning 0

	# Memory channels.
	for {set bn 0} {$bn < 10} {incr bn} \
		{
		set ch [expr {$bn * 100}]
		for {set i 0} {$i < 64} {incr i} \
			{
			set m "Ch. $ch"
			set f $MemFreq($ch)

			set code [ValidateFreq $f $m]
			if {$code == "error"} { incr nerror } \
			elseif {$code == "warning"} { incr nwarning }
	
			set f $MemMode($ch)
			set code [ValidateMode $f $m]
			if {$code == "error"} { incr nerror }
	
			if { [expr {$nerror + $nwarning}] > 5} {break}
			incr ch
			}
		}

#	puts stderr "ValidateData: done with memories"


	# WX freq
	set f $GlobalParam(WXFreq)
	set m "WX"
	set code [ValidateFreq $f $m]
	if {$code == "error"} { incr nerror } \
	elseif {$code == "warning"} { incr nwarning }

	set f $GlobalParam(WXMode)
	set code [ValidateMode $f $m]
	if {$code == "error"} { incr nerror }



	# VFO
	set f $GlobalParam(VFOFreq)
	set m "VFO"
	set code [ValidateFreq $f $m]
	if {$code == "error"} { incr nerror } \
	elseif {$code == "warning"} { incr nwarning }

	set f $GlobalParam(VFOMode)
	set code [ValidateMode $f $m]
	if {$code == "error"} { incr nerror }

	# Priority channel
	set f $Priority
	set m "Priority"
	set code [ValidateFreq $f $m]
	if {$code == "error"} { incr nerror } \
	elseif {$code == "warning"} { incr nwarning }


	set f $PriorityMode
	set code [ValidateMode $f $m]
	if {$code == "error"} { incr nerror }


	# Smart Search freqs
	set f $SmartSearch(lower)
	set m "SmartSearch"
	set code [ValidateFreq $f $m]
	if {$code == "error"} { incr nerror } \
	elseif {$code == "warning"} { incr nwarning }

	set f $SmartSearch(upper)
	set m "SmartSearch"
	set code [ValidateFreq $f $m]
	if {$code == "error"} { incr nerror } \
	elseif {$code == "warning"} { incr nwarning }

	set f $SmartSearch(start)
	set m "SmartSearch"
	set code [ValidateFreq $f $m]
	if {$code == "error"} { incr nerror } \
	elseif {$code == "warning"} { incr nwarning }

	# Limit scan
	for {set i 0} {$i < 9} {incr i} \
		{
		set m "Limit Scan bank L0$i"
		set f $LimitScan($i,lower)

		set code [ValidateFreq $f $m]
		if {$code == "error"} { incr nerror } \
		elseif {$code == "warning"} { incr nwarning }

		set m "Limit Scan bank L0$i"
		set f $LimitScan($i,upper)

		set code [ValidateFreq $f $m]
		if {$code == "error"} { incr nerror } \
		elseif {$code == "warning"} { incr nwarning }

		if { [expr {$nerror + $nwarning}] > 5} {break}
		}


	# Skip frequencies
	for {set i 0} {$i < 64} {incr i} \
		{
		set m "Skip frequency S$i"
		set f $Skip($i,freq)

		set code [ValidateFreq $f $m]
		if {$code == "error"} { incr nerror } \
		elseif {$code == "warning"} { incr nwarning }

		if { [expr {$nerror + $nwarning}] > 5} {break}
		}


	# Band Edge frequencies
	for {set i 0} {$i < 12} {incr i} \
		{
		set m "Band Edge frequency $i"
		set f $BandEdge($i,freq)

		set code [ValidateFreq $f $m]
		if {$code == "error"} { incr nerror } \
		elseif {$code == "warning"} { incr nwarning }

		if { [expr {$nerror + $nwarning}] > 5} {break}
		}


	# BandStack frequencies
	for {set i 0} {$i < 12} {incr i} \
		{
		set m "Band Stacking frequency $i"
		set f $BandStack($i,freq)

		set code [ValidateFreq $f $m]
		if {$code == "error"} { incr nerror } \
		elseif {$code == "warning"} { incr nwarning }

		if { [expr {$nerror + $nwarning}] > 5} {break}
		}

	set code [ValidateBS]
	if {$code == "error"} { incr nerror } \
	elseif {$code == "warning"} { incr nwarning }

	# Dual Watch banks
	for {set i 0} {$i < 9} {incr i} \
		{
		set m "Dual Watch pair A0$i"

		set f $DualWatch($i,first)
		set code [ValidateFreq $f $m]
		if {$code == "error"} { incr nerror} \
		elseif {$code == "warning"} { incr nwarning }

		set f $DualWatch($i,mode1)
		set code [ValidateMode $f $m]
		if {$code == "error"} { incr nerror }

		set m "Dual Watch pair $i"

		set f $DualWatch($i,second)
		set code [ValidateFreq $f $m]
		if {$code == "error"} { incr nerror } \
		elseif {$code == "warning"} { incr nwarning }

		set f $DualWatch($i,mode2)
		set code [ValidateMode $f $m]
		if {$code == "error"} { incr nerror }

		if { [expr {$nerror + $nwarning}] > 5} {break}
		}


	# Preset frequencies
	for {set i 0} {$i < 10} {incr i} \
		{
		set m "Preset $i"

		set f $Preset($i,freq)
		set code [ValidateFreq $f $m]
		if {$code == "error"} { incr nerror } \
		elseif {$code == "warning"} { incr nwarning }

		set f $Preset($i,mode)
		set code [ValidateMode $f $m]
		if {$code == "error"} { incr nerror }

		if { [expr {$nerror + $nwarning}] > 5} {break}
		}


	# Auto band
	for {set i 1} {$i < 30} {incr i} \
		{
		set m "Auto Band $i"

		set f $AutoBand($i,freq)
		set code [ValidateFreq $f $m]
		if {$code == "error"} { incr nerror } \
		elseif {$code == "warning"} { incr nwarning }

		set f $AutoBand($i,mode)
		set code [ValidateMode $f $m]
		if {$code == "error"} { incr nerror }

		if { [expr {$nerror + $nwarning}] > 5} {break}
		}

	set code [ValidateTweaks]
	if {$code == "error"} { incr nerror }

	if {$nerror} \
		{
		tk_dialog .baddata1 "tk120 Invalid data" \
			$Emsg error 0 OK
		# puts stderr "ValidateData: returning 1"
		return 1
		}
	if {$nwarning} \
		{
		set response [tk_dialog .baddata2 \
			"tk120 Data warning" \
			$Emsg error 0 Cancel Continue]

		if {$response == 0} then {return 1} \
		else {return 0}
		}
	
	return 0
}
######################################################
# Check to make sure the frequency in each
# bandstacking memory is within the frequency limits
# of the corresponding band.
#
# Returns:
#	""	-ok
#	error	-if there are one or more errors
#
# Note:
#	Appends error message to the global variable Emsg
######################################################
proc ValidateBS { } \
{
	global BandEdge
	global BandStack
	global Emsg

	set code ""
	set msg ""

	for {set i 0} {$i < 12} {incr i} \
		{
		set f $BandStack($i,freq)

		if { $f > $BandEdge($i,freq) } \
			{
			set code error
			append msg "Bandstacking memory $i ($f MHz) "
			append msg "is out of band.\n"
			}

		set j $i
		incr j -1

		if {$i > 0} \
			{
			if {$f < $BandEdge($j,freq)} \
				{
				set code error
				append msg "Bandstacking memory $i ($f)"
				append msg "is out of band.\n"
				}
			}
		}
	append Emsg $msg
	return $code
}

###################################################################
# Check a frequency for validity.
# Append the error or warning message to a global string.
#
# Returns:
#	0
#	warning
#	error
###################################################################
proc ValidateFreq {f m} \
{
	global Emsg
	global GlobalParam

	set code 0
	set msg ""

	if {( ($f != "") && ($f != 0.0) ) \
		&& (($f < .1000) || ($f > 1300.0))} \
		{
		append msg "\nError: $m frequency ($f) is out"
		append msg " of range.\n"
		set code error
		}

	if { $GlobalParam(WhichModel) == "USA" } \
		{
		# Check for censored freqs in USA model.

		if { ($f >= 327.0) && ($f < 328.0)} \
			{
			append msg "\nWarning: $m frequency ($f) is"
			append msg " blocked in the US version.\n"
			set code warning
			} \
		elseif { ($f >= 339.0) && ($f < 340.0)} \
			{
			append msg "\nWarning: $m frequency ($f) is"
			append msg " blocked in the US version.\n"
			set code warning
			} \
		elseif { ($f >= 351.0) && ($f < 352.0)} \
			{
			append msg "\nWarning: $m frequency ($f) is"
			append msg " blocked in the US version.\n"
			set code warning
			} \
		elseif {($f >= 372.0) && ($f < 373.0)} \
			{
			append msg "\nWarning: $m frequency ($f) is"
			append msg " blocked in the US version.\n"
			set code warning
			} \
		elseif {($f >= 384.0) && ($f < 385.0)} \
			{
			append msg "\nWarning: $m frequency ($f) is"
			append msg " blocked in the US version.\n"
			set code warning
			} \
		elseif {($f >= 396.0) && ($f < 397.0)} \
			{
			append msg "\nWarning: $m frequency ($f) is"
			append msg " blocked in the US version.\n"
			set code warning
			} \
		elseif {($f >= 474.0) && ($f < 475.0)} \
			{
			append msg "\nWarning: $m frequency ($f) is"
			append msg " blocked in the US version.\n"
			set code warning
			} \
		elseif {($f >= 482.0) && ($f < 483.0)} \
			{
			append msg "\nWarning: $m frequency ($f) is"
			append msg " blocked in the US version.\n"
			set code warning
			} \
		elseif {($f >= 496.0) && ($f < 497.0)} \
			{
			append msg "\nWarning: $m frequency ($f) is"
			append msg " blocked in the US version.\n"
			set code warning
			} \
		elseif {($f >= 504.0) && ($f < 505.0)} \
			{
			append msg "\nWarning: $m frequency ($f) is"
			append msg " blocked in the US version.\n"
			set code warning
			} \
		elseif {($f >= 511.0) && ($f < 512.0)} \
			{
			append msg "\nWarning: $m frequency ($f) is"
			append msg " blocked in the US version.\n"
			set code warning
			} \
		elseif {($f >= 542.0) && ($f < 549.0)} \
			{
			append msg "\nWarning: $m frequency ($f) is"
			append msg " blocked in the US version.\n"
			set code warning
			} \
		elseif {($f >= 558.0) && ($f < 572.0)} \
			{
			append msg "\nWarning: $m frequency ($f) is"
			append msg " blocked in the US version.\n"
			set code warning
			} \
		elseif {($f >= 605.0) && ($f < 615.0)} \
			{
			append msg "\nWarning: $m frequency ($f) is"
			append msg " blocked in the US version.\n"
			set code warning
			} \
		elseif {($f >= 620.0) && ($f < 630.0)} \
			{
			append msg "\nWarning: $m frequency ($f) is"
			append msg " blocked in the US version.\n"
			set code warning
			} \
		elseif {($f >= 784.0) && ($f < 798.0)} \
			{
			append msg "\nWarning: $m frequency ($f) is"
			append msg " blocked in the US version.\n"
			set code warning
			} \
		elseif {($f >= 807.0) && ($f < 820.0)} \
			{
			append msg "\nWarning: $m frequency ($f) is"
			append msg " blocked in the US version.\n"
			set code warning
			} \
		elseif {($f >= 869.0) && ($f < 894.0)} \
			{
			append msg "\nWarning: $m frequency ($f) is"
			append msg " blocked in the US version.\n"
			set code warning
			} \
		elseif {($f >= 824.0) && ($f < 849.0)} \
			{
			append msg "\nWarning: $m frequency ($f) is"
			append msg " blocked in the US version.\n"
			set code warning
			} \
		elseif {($f >= 1001.0) && ($f < 1002.0)} \
			{
			append msg "\nWarning: $m frequency ($f) is"
			append msg " blocked in the US version.\n"
			set code warning
			} \
		elseif {($f >= 1009.0) && ($f < 1010.0)} \
			{
			append msg "\nWarning: $m frequency ($f) is"
			append msg " blocked in the US version.\n"
			set code warning
			}
		}

	append Emsg $msg
	return $code
}

###################################################################
# Check that Tweak fields are valid.
###################################################################
proc ValidateTweaks { } \
{
	global Emsg
	global GlobalParam

	set code 0

	foreach addr {1A0C 1A0E 1A10} \
		{
		if { [IsHex $GlobalParam($addr)] == 0 } \
			{
			append Emsg "\nError: $addr must be a "
			append Emsg "hex value (00 - ff) "
			set code error
			}
		}


	return $code
}

###################################################################
# Return 1 if a string consists of 2 hex digits.
###################################################################
proc IsHex { s } \
{
	# Check for non-digit and non decimal point chars.
	set rc [regexp -nocase {^[0-9a-f][0-9a-f]$} $s]
	if {$rc} \
		{
		return 1
		} \
	else \
		{
		return 0
		}
}

###################################################################
# Check a mode for validity.
# Append the error message to a global string.
#
# Returns:
#	0
#	warning
#	error
###################################################################
proc ValidateMode {mode m} \
{
	global Emsg
	global Mode

	set code 0

	if { [info exists Mode($mode)] == 0} \
		{
		append Emsg "\nError: $m mode ($mode) is invalid.\n"
		set code error
		} 

	return $code
}

###################################################################
# Set title of the main window so it contains the
# current template file name.
###################################################################
proc SetWinTitle { } \
{
	global GlobalParam

	if { ( [info exists GlobalParam(TemplateFilename)] == 0 ) \
		|| ($GlobalParam(TemplateFilename) == "") } \
		{
		set filename untitled.tk1
		} \
	else \
		{
		set filename $GlobalParam(TemplateFilename)
		}

	set s [format "%s - tk120" $filename]
	wm title . $s

	return
}

###################################################################
# Given a frequency, mode, and preset, return the 4 byte
# encoded internal representation that Yaesu employs.
###################################################################
proc FreqMode2BCD { f mode pref } \
{
	global Mode


	if {$f < .001} \
		{
		# Frequency is zero MHz so set all bits.
		set s [binary format "H2H2H2H2" FF FF FF FF]
		return $s
		}
		
	set freqdigits [Freq2BCD3p $f]

	binary scan $freqdigits "H*" junk

	set splinter [string index $freqdigits 3]
	set freqdigits [string range $freqdigits 0 2]

	set mode [string toupper $mode]
	if {[info exists Mode($mode)]} \
		{
		set nmode $Mode($mode)
		} \
	else \
		{
		set nmode $Mode(NFM)
		}

	set m [format "0%1d" $nmode]
	set b [binary format "H2" $m]

	if { $pref == "p" } \
		{
		# Preferential scan channel
		set b [SetBit $b 5]
		}

	binary scan $splinter "H2" spl
	if { $spl == "40" } \
		{
		# A 12.5 kHz splinter freq.
		set b [SetBit $b 1]
		} \
	elseif { $spl == "20" } \
		{
		# A 6.25 kHz splinter freq.
		set b [SetBit $b 2]
		}

	append freqdigits $b

	return $freqdigits
}

# Prevent user from shrinking or expanding window.

proc FixSize { w } \
{
	wm minsize $w [winfo width $w] [winfo height $w]
	wm maxsize $w [winfo width $w] [winfo height $w]

	return
}

proc DecodeFlagByte3 { image } \
{
	global GlobalParam
	global ImageAddr

	# Parse 1A11
	scan $ImageAddr(FlagByte3) "%x" first
	set s [string range $image $first $first]
	set t [Char2Int $s]
	set GlobalParam(1A11) [format "%02x" $t]
	update idletasks

	return
}



######################################################################
#					Bob Parnass
#					DATE:
#
# USAGE:	SortaBank first last
#
# INPUTS:
#		first	-starting channel to sort
#		last	-ending channel to sort
#
# RETURNS:
#		0	-ok
#		-1	-error
#
#
# PURPOSE:	Sort a range of memory channels based on frequency.
#
# DESCRIPTION:
#
######################################################################
proc SortaBank { first last } \
{
	global GlobalParam
	global MemFreq
	global MemLabel
	global MemMode
	global MemPref
	global Mimage


	if {[info exists Mimage] == 0} \
		{
		set msg "You must open a template file\n"
		append msg " or read an image from the radio\n"
		append msg " before sorting channels.\n"

		tk_dialog .belch "tk120" \
			$msg info 0 OK
		return -1
		}

	if {$GlobalParam(SortType) == "freq"} \
		{
		set inlist [Bank2List MemFreq $first $last]
		set vorder [SortFreqList $inlist]
		} \
	else \
		{
		set inlist [Bank2List MemLabel $first $last]
		set vorder [SortLabelList $inlist]
		}


	set inlist [Bank2List MemFreq $first $last]
	set slist [ReorderList $inlist $vorder]
	for {set i $first; set j 0} {$i <= $last} {incr i; incr j} \
		{
		set MemFreq($i) [lindex $slist $j]
		}

	set inlist [Bank2List MemMode $first $last]
	set slist [ReorderList $inlist $vorder]
	for {set i $first; set j 0} {$i <= $last} {incr i; incr j} \
		{
		set MemMode($i) [lindex $slist $j]
		if {$MemMode($i) == ""} \
			{
			set MemMode($i) NFM
			}
		}

	set inlist [Bank2List MemLabel $first $last]
	set slist [ReorderList $inlist $vorder]
	for {set i $first; set j 0} {$i <= $last} {incr i; incr j} \
		{
		set MemLabel($i) [lindex $slist $j]
		}

	set inlist [Bank2List MemPref $first $last]
	set slist [ReorderList $inlist $vorder]
	for {set i $first; set j 0} {$i <= $last} {incr i; incr j} \
		{
		set MemPref($i) [lindex $slist $j]
		}

	return 0
}


###################################################################
#
# Clear all memory channels by zeroing the proper arrays,
# but confirm this with the user first.
#
###################################################################
proc ClearAllChannels { } \
{
	global GlobalParam
	global Mimage
	global MemFreq
	global MemLabel
	global MemMode
	global MemPref


	if { ([info exists Mimage] == 0) \
		|| ([string length $Mimage] <= 0)} \
		{
		# No image to write.
		set msg "You must first read template data from"
		append msg " the radio or open a file before"
		append msg " clearing memories."

		tk_dialog .error "Clear all channels" $msg error 0 OK
		return
		}


	set msg "Warning: This operation will clear all "
	append msg "memory channels."

	set result [tk_dialog .clearall "Warning" \
		$msg warning 0 Cancel "Clear Memories" ]

	if {$result == 0} {return}


	for {set bn 0} {$bn < 10} {incr bn} \
		{
		set ch [expr {$bn * 100}]
		for {set i 0} {$i < 64} {incr i} \
			{
			set MemFreq($ch) 0
			set MemLabel($ch) ""
			set MemMode($ch) NFM
			set MemPref($ch) ""
			incr ch
			}
		}
	return

}


###################################################################
#
# Parse data from inside the memory image and store it
# in global arrays.
#
###################################################################
proc DecodeScanResume { } \
{
	global GlobalParam
	global ImageAddr
	global Mimage
	global RScanResume


	# Parse VFO scan resume mode.
	scan $ImageAddr(ScanResumeVFO) "%x" first
	set s [string index $Mimage $first ]

	binary scan $s "H2" s
	set m [string index $s 1]
	if { [info exists RScanResume($m)] } \
		{
		set GlobalParam(ScanResumeVFO) \
			[string toupper $RScanResume($m)]
		} \
	else \
		{
		set GlobalParam(ScanResumeVFO) BUSY
		}


	# Parse Memory scan resume mode.
	scan $ImageAddr(ScanResumeMem) "%x" first
	set s [string index $Mimage $first ]

	binary scan $s "H2" s
	set m [string index $s 1]
	if { [info exists RScanResume($m)] } \
		{
		set GlobalParam(ScanResumeMem) \
			[string toupper $RScanResume($m)]
		} \
	else \
		{
		set GlobalParam(ScanResumeMem) BUSY
		}


	# Parse Dual Watch scan resume mode.
	scan $ImageAddr(ScanResumeDW) "%x" first
	set s [string index $Mimage $first ]

	binary scan $s "H2" s
	set m [string index $s 1]
	if { [info exists RScanResume($m)] } \
		{
		set GlobalParam(ScanResumeDW) \
			[string toupper $RScanResume($m)]
		} \
	else \
		{
		set GlobalParam(ScanResumeDW) BUSY
		}

	return
}

######################################################
# Determine the default mode for a given frequency
# based on the Auto Band plan.
#
# Returns:
#	mode
#
######################################################
proc DefaultMode { f } \
{
	global AutoBand

	set mode AM

	set j 2
	for {set i 1} {$i < 40} {incr i} \
		{
		set mode $AutoBand($i,mode)

		if { $AutoBand($j,freq) < .001} \
			{
			break
			}

		if { ($f >= $AutoBand($i,freq)) \
			&& ($f < $AutoBand($j,freq)) } \
			{
			set mode $AutoBand($i,mode)
			break
			}
		incr j
		}
	# puts stderr "DefaultMode: f: $f, mode: $mode, i: $i"
	return $mode
}


###################################################################
# Encode Scan Resume 
# information into a memory image.
###################################################################
proc EncodeScanResume { image } \
{
	global ImageAddr
	global GlobalParam
	global Mimage
	global ScanResume


	# encode VFO Scan Resume.
	scan $ImageAddr(ScanResumeVFO) "%x" first
	set c $GlobalParam(ScanResumeVFO)

	if { [info exists ScanResume($c)] } \
		{
		set c $ScanResume($c)
		} \
	else { set c $ScanResume(BUSY) }

	set s [format "%02x" $c]
	set b [binary format "H2" $s]
	set image [string replace $image $first $first $b]

	# encode Memory Scan Resume.
	scan $ImageAddr(ScanResumeMem) "%x" first
	set c $GlobalParam(ScanResumeMem)

	if { [info exists ScanResume($c)] } \
		{
		set c $ScanResume($c)
		} \
	else { set c $ScanResume(BUSY) }

	set s [format "%02x" $c]
	set b [binary format "H2" $s]
	set image [string replace $image $first $first $b]

	# encode Dual Watch Scan Resume.
	scan $ImageAddr(ScanResumeDW) "%x" first
	set c $GlobalParam(ScanResumeDW)

	if { [info exists ScanResume($c)] } \
		{
		set c $ScanResume($c)
		} \
	else { set c $ScanResume(BUSY) }

	set s [format "%02x" $c]
	set b [binary format "H2" $s]
	set image [string replace $image $first $first $b]


	return $image
}


###################################################################
# Create 10 memory banks. 
###################################################################
proc MakeMemoryBankFrame { f }\
{
	global GlobalParam
	global MemNB

	frame $f -relief groove -borderwidth 3
	# frame $f.b -relief flat -borderwidth 3

	label $f.lab -text "Memory Bank Settings" \
		-borderwidth 3


	pack $f.lab \
		-side top -fill both -expand false -padx 3 -pady 3

	MakeMemoryBankFrameCommon $f.common
	pack $f.common  -side left -padx 3 -pady 3 \
		-fill both

	set MemNB $f.banknb
	MakeMemoryBankNB $MemNB
	return $f
}

###################################################################
# Make a frame for settings common to all memory banks.
###################################################################
proc MakeMemoryBankFrameCommon { f }\
{
	global GlobalParam

	frame $f -relief groove -borderwidth 3

	set t $f.t
	frame $t -relief groove -borderwidth 3

	label $t.lbs -text "Bank Scan" -borderwidth 3
	checkbutton $t.bs -text "" \
		-variable GlobalParam(BankScan) \
		-onvalue 1 -offvalue 0


	set hint ""
	set hint [append hint "Bank Scan restricts the radio "]
	set hint [append hint "to scan a channel bank or "]
	set hint [append hint "a set of channel banks (choose below) "]
	set hint [append hint "instead of scanning all banks. "]
	balloonhelp_for $t.bs $hint
	balloonhelp_for $t.lbs $hint


        label $t.lsresume -text "Memory\nScan\nResume" -borderwidth 3
	tk_optionMenu $t.sresume GlobalParam(ScanResumeMem) \
		PAUSE BUSY HOLD

	grid $t.lbs -row 5 -column 0 -sticky w
	grid $t.bs -row 5 -column 1 -sticky w
	grid $t.lsresume -row 10 -column 0 -sticky w
	grid $t.sresume -row 10 -column 1 -sticky w

	pack $t -side top -padx 3 -pady 3

	set l [MakeLinkedScanFrame $f.banklinks]
	pack $l -side top -padx 3 -pady 3 -expand true -fill both
	return $f
}

###################################################################
# Make a single tabbed notebook to hold the settings
# for 10 individual memory banks.
###################################################################
proc MakeMemoryBankNB { w } \
{
	global NBanks

	tabnotebook_create $w

	for {set i 0} {$i < 10} {incr i} \
		{
		set p [tabnotebook_page $w "Bank $i"]
		set fr $p.f; MakeMemoryBankPage $fr $i
		pack $fr
		}

	pack $w -expand true -fill both
}

###################################################################
# Make a frame to hold the settings
# for one memory bank.
###################################################################
proc MakeMemoryBankPage { f bn }\
{
	global GlobalParam

	frame $f -relief flat -borderwidth 3

	MakeBankSettingsFrame $f.c1 $bn
	MakeMemoryChannelFrame $f.c2 $bn
	MakeFillerFrame $f.c3 $bn

	pack $f.c1 $f.c2 $f.c3 -side left \
		-fill both -padx 3 -pady 3 -expand true

	return $f
}


###################################################################
# Create one a set of widgets which pertain to
# for one memory bank. 
#
# INPUTS:
#		f	-frame to create
#		bn	-bank number
#
###################################################################
proc MakeBankSettingsFrame { f bn }\
{
	global DisplayFontSize 

	frame $f -relief flat -borderwidth 3


	label $f.bankn$bn -text "$bn" \
		-relief flat \
		-borderwidth 6 \
		-font $DisplayFontSize 


	grid $f.bankn$bn -row 5 -column 0 -sticky ew 

	return $f
}


###################################################################
#
# INPUTS:
#		f	-frame to create
#		bn	-bank number
#
###################################################################
proc MakeFillerFrame { f bn }\
{
	global DisplayFontSize 

	frame $f -relief flat -borderwidth 3


	for {set i 0} {$i < 12} {incr i} \
		{
		label $f.filler$bn$i -text "-"  -relief flat \
			-borderwidth 6

		grid $f.filler$bn$i -row $i -column 0 -sticky ew 
		}

	return $f
}


###################################################################
# Create widgets for memory channels for a bank. 
###################################################################
proc MakeMemoryChannelFrame { f bn }\
{
	global NBanks
	global NChanPerBank
	global VNChanPerBank

	frame $f -relief flat -borderwidth 3
	label $f.lab -text "Memory Channels" -borderwidth 3

	pack $f.lab -side top


	ScrollformCreate $f.b
	pack $f.b -expand yes -fill both

	set w [ScrollFormInterior $f.b]


	label $w.freq -text "Freq" -borderwidth 3
	label $w.mode -text "Mode" -borderwidth 3
	label $w.lab -text "Label" -borderwidth 3
	label $w.pref -text "Pref" -borderwidth 3

	set ch [expr {$bn * $VNChanPerBank}]
	for {set i 0} {$i < $NChanPerBank} {incr i} \
		{
		MakeChannel $w $bn $i $ch
		incr ch
		update idletasks
		}

	grid $w.freq -row 0 -column 20
	grid $w.mode -row 0 -column 30
	grid $w.lab -row 0 -column 40
	grid $w.pref -row 0 -column 50

	return $f
}


###################################################################
# Create one a set of widgets for one skip talk group. 
###################################################################
proc MakeChannel { f bn n ch }\
{
	global GlobalParam
	global MemFreq
	global MemLabel
	global MemMode
	global MemPref

	set row [expr {$n + 2}]


	label $f.lab$bn$ch -text "$ch" -borderwidth 3

	entry $f.freq$bn$ch -width 12 \
		-textvariable MemFreq($ch) \
		-background white 

	entry $f.label$bn$ch -width 10 \
		-textvariable MemLabel($ch) \
		-background white 

	tk_optionMenu $f.modemenu$bn$ch MemMode($ch) \
		NFM WFM AM


	checkbutton $f.pref$bn$ch -text "" \
		-variable MemPref($ch) \
		-onvalue p \
		-offvalue ""



	button $f.lower$bn$ch -text "^" \
		-command "SwapChannel $ch [expr {$ch - 1}] 1"

	button $f.higher$bn$ch -text "v" \
		-command "SwapChannel $ch [expr { $ch + 1}] 1"

	button $f.insert$bn$ch -text "Insert" \
		-command "InsertChannel $ch"

	button $f.delete$bn$ch -text "Delete" \
		-command "DeleteChannel $ch"

	grid $f.lab$bn$ch -row $row -column 10
	grid $f.freq$bn$ch -row $row -column 20
	grid $f.modemenu$bn$ch -row $row -column 30 -sticky w
	grid $f.label$bn$ch -row $row -column 40
	grid $f.pref$bn$ch -row $row -column 50
	grid $f.lower$bn$ch -row $row -column 60
	grid $f.higher$bn$ch -row $row -column 70
	grid $f.insert$bn$ch -row $row -column 80
	grid $f.delete$bn$ch -row $row -column 90

	return
}

###################################################################
# Insert a memory channel and move all the higher channels
# in the same bank higher by one channel.  Clear the current
# channel in the bank.
###################################################################
proc InsertChannel { ch } \
{
	global MemFreq
	global MemLabel
	global MemMode
	global MemPref
	global NChanPerBank
	global VNChanPerBank

	set bn [expr {int($ch/$VNChanPerBank)}]
	set last [expr {($bn * $VNChanPerBank) + $NChanPerBank - 1}]

	if {$MemFreq($last) > 0} \
		{
		# No room.

		set msg "Channel $last is not empty.\n\n"
		append msg "Please delete channel $last before "
		append msg "inserting a new channel $ch and "
		append msg "moving the existing channels higher."

		tk_dialog .belch "Insert new channel" \
			$msg error 0 OK

		return
		}

	set n [expr {$NChanPerBank - fmod($ch, $VNChanPerBank) - 1}]

	set to $last
	set from $last
	incr from -1

	for {set i 0} {$i < $n} {incr i} \
		{
		# puts stderr "InsertChannel: n: $n, moving channel $from to $to"
		set MemFreq($to) $MemFreq($from)
		set MemLabel($to) $MemLabel($from)
		set MemMode($to) $MemMode($from)
		set MemPref($to) $MemPref($from)

		incr from -1
		incr to	-1
		}

	set MemFreq($ch) 0.00000
	set MemLabel($ch) ""
	set MemMode($ch) NFM
	set MemPref($ch) ""

}

###################################################################
# Delete a memory channel and move all the higher channels
# in the same bank to the previous channel.  Clear the last
# channel in the bank.
###################################################################

proc DeleteChannel { ch } \
{
	global MemFreq
	global MemLabel
	global MemMode
	global MemPref
	global NChanPerBank
	global VNChanPerBank

	set n [expr {$NChanPerBank - fmod($ch, $VNChanPerBank) - 1}]

	set to $ch
	set from $ch
	incr from

	for {set i 0} {$i < $n} {incr i} \
		{
		# puts stderr "DeleteChannel: moving channel $from to $to"
		set MemFreq($to) $MemFreq($from)
		set MemLabel($to) $MemLabel($from)
		set MemMode($to) $MemMode($from)
		set MemPref($to) $MemPref($from)

		incr from	
		incr to	
		}

	set bn [expr {int($ch/$VNChanPerBank)}]
	set last [expr {($bn * $VNChanPerBank) + $NChanPerBank - 1}]

	set MemFreq($last) 0.00000
	set MemLabel($last) ""
	set MemMode($last) NFM
	set MemPref($last) ""

	return
}


###################################################################
# Swap channel with the another channel in the bank.
#
# INPUTS:
#		ch1		-first channel
#		ch2		-second channel
#		samebank	-1 = channels must be withn
#				the same bank.
#				0 = channels may be in different
#				banks.
###################################################################
proc SwapChannel { ch1 ch2 samebank } \
{
	global MemFreq
	global MemLabel
	global MemMode
	global MemPref
	global NChanPerBank
	global VNChanPerBank

	set bn [expr {int($ch1/$VNChanPerBank)}]
	set last [expr {($bn * $VNChanPerBank) + $NChanPerBank - 1}]
	set ch1r [expr {int(fmod($ch1,$VNChanPerBank))}]
	set ch2r [expr {int(fmod($ch2,$VNChanPerBank))}]

	if { $samebank && ( ($ch1 > $last) \
		|| ($ch1 < 0) \
		|| ($ch2 > $last) \
		|| ($ch2 < 0) \
		|| ($ch1r >= $NChanPerBank) \
		|| ($ch2r >= $NChanPerBank) ) } \
		{

		set msg "Cannot move channel $ch1 to $ch2."

		tk_dialog .belch "Move channel" \
			$msg error 0 OK

		return
		}

	set tmp $MemFreq($ch2)
	set MemFreq($ch2) $MemFreq($ch1)
	set MemFreq($ch1) $tmp

	set tmp $MemLabel($ch2)
	set MemLabel($ch2) $MemLabel($ch1)
	set MemLabel($ch1) $tmp

	set tmp $MemMode($ch2)
	set MemMode($ch2) $MemMode($ch1)
	set MemMode($ch1) $tmp

	set tmp $MemPref($ch2)
	set MemPref($ch2) $MemPref($ch1)
	set MemPref($ch1) $tmp

	return
}


###################################################################
#
# Swap the bank settings, memory channel info, fleet map,
# and talk group info for 2 banks with each other.
#
###################################################################

proc SwapBank { bn1 bn2 } \
{
	global ChanBank
	global GlobalParam
	global NBanks
	global NChanPerBank
	global VNChanPerBank

	# Sanity checks.
	if { $bn1 == $bn2 } \
		{
		# Swap bank with itself.
		return
		}

	if { ($bn1 < 0) || ($bn1 >= $NBanks) \
		|| ($bn2 < 0) || ($bn2 >= $NBanks) } \
		{
		# No such bank.
		return
		}

	# Swap the memory channels.
	set ch1 [expr {$bn1 * $VNChanPerBank}]
	set ch2 [expr {$bn2 * $VNChanPerBank}]

	for {set i 0} {$i < $NChanPerBank} {incr i} \
		{
		SwapChannel $ch1 $ch2 0
		incr ch1
		incr ch2
		}

	return
}
